#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 21:39:40 2015

@author: Brandon Logan and Gene Callahan

Segregation Run File
"""

import indra.utils as utils
import indra.prop_args as props
import schelling.segregation_model as sm

# set up some file names:
MODEL_NM = "segregation_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 60)
        utils.get_agent_num(pa, "num_R_agents", "red agents", 1100)
        utils.get_agent_num(pa, "num_B_agents", "blue agents", 1100)
        utils.get_pct(pa, "max_tolerance", "agent", "minimum intolerance", .10)
        utils.get_pct(pa, "min_tolerance", "agent", "maximum intolerance", .70)
        pa.ask("max_detect", "What is the agent's neighborhood size?",
               int, default=4, limits=(1, 10))
    # Now we create an environment for our agents to act within:
    env = sm.SegregationEnv("A city",
                            pa.get("grid_width"),
                            pa.get("grid_height"),
                            props=pa)
    # Now we loop creating multiple agents with numbered names
    # based on the loop variable:
    for i in range(pa.get("num_B_agents")):
        env.add_agent(sm.BlueAgent(name="Blue agent" + str(i),
                      goal="A good neighborhood.",
                      min_tol=pa.get('min_tolerance'),
                      max_tol=pa.get('max_tolerance'),
                      max_detect=pa.get('max_detect')))
    for i in range(pa.get("num_R_agents")):
        env.add_agent(sm.RedAgent(name="Red agent" + str(i),
                      goal="A good neighborhood.",
                      min_tol=pa.get('min_tolerance'),
                      max_tol=pa.get('max_tolerance'),
                      max_detect=pa.get('max_detect')))
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
vector_space.py
The way agents interact is through prehensions.
This implements prehensions as vector spaces.
"""

import math
import numpy as np
import indra.prehension as pre
# import logging

# x and y indices
X = 0
Y = 1

# Set up constants for some common vectors: this will save time and memory.
X_VEC = np.array([1, 0])
Y_VEC = np.array([0, 1])
NULL_VEC = np.array([0, 0])
NEUT_VEC = np.array([.7071068, .7071068])


def from_vector(v):
    """
    Convenience method to turn a vector into a prehension.
    """
    vspace = VectorSpace()
    vspace.vector = v
    return vspace


def normalize(v):
    norm = np.linalg.norm(v)
    if norm == 0:
        return v
    else:
        return v / norm


class VectorSpace(pre.Prehension):
    """
    We need to put definition of a vector space here.
    """

# we must pre-declare these, then use them, then init them at the bottom
#  of the file.
    X_PRE = None
    Y_PRE = None
    NULL_PRE = None
    NEUT_PRE = None


    def __init__(self, x=0, y=0):
        super().__init__()
        self.vector = np.array([x, y])

    def __str__(self):
        return ("x: %f, y: %f" % (self.vector[X], self.vector[Y]))

    def prehend(self, other):
        """
        In this class, a prehension prehends another prehension
        through vector addition.
        other: prehension to prehend
        """
        return from_vector(self.vector + other.vector)

    def intensify(self, a):
        """
        Here this is scalar multiplication of a vector.
        a: scalar to multiply by.
        """
        return from_vector(self.vector * a)

    def direction(self):
        """
        This gets us the orientation of the vector: x, y, or neutral.
        We use it, for instance, to set an agent's market stance to
        buy or sell.
        """
        if self.vector[X] > self.vector[Y]:
            return VectorSpace.X_PRE
        elif self.vector[X] < self.vector[Y]:
            return VectorSpace.Y_PRE
        else:
            return VectorSpace.NEUT_PRE

    def project(self, x_or_y):
        """
        Projects the vector onto the x or y axis.
        Pass in X or Y as declared above.
        """
        return self.vector[x_or_y]

    def equals(self, other):
        """
        For prehensions of the base type, they are equal
        when their vetors are equal.
        """
        return np.array_equal(self.vector, other.vector)

    def reverse(self):
        """
        Reverse the vector.
        Reflection across line y = x.
        """
        new_vec = np.array(np.flipud(self.vector))
        return from_vector(new_vec)

    def normalize(self):
        """
        Return a normalized prehension.
        If we get the NULL prehension, just return it.
        """
        if self.equals(VectorSpace.NULL_PRE):
            return VectorSpace.NULL_PRE
        else:
            return from_vector(normalize(self.vector))


# Now we actually initialize the prehensions we declared above.
#  This can't be done earlier, since VectorSpace was just defined.
VectorSpace.X_PRE = from_vector(X_VEC)
VectorSpace.Y_PRE = from_vector(Y_VEC)
VectorSpace.NULL_PRE = from_vector(NULL_VEC)
VectorSpace.NEUT_PRE = from_vector(NEUT_VEC)


def stance_pct_to_pre(pct, x_or_y):
    """
    pct is our % of the way to the y-axis from
    the x-axis around the unit circle. (If x_or_y == Y, it is the opposite.)
    It will return the x, y coordinates of the point that % of the way.
    I.e., .5 returns NEUT_VEC, 0 returns X_VEC.
    """
    if x_or_y == Y:
        pct = 1 - pct
    if pct == 0:
        return VectorSpace.X_PRE
    elif pct == .5:
        return VectorSpace.NEUT_PRE
    elif pct == 1:
        return VectorSpace.Y_PRE
    else:
        angle = 90 * pct
        x = math.cos(math.radians(angle))
        y = math.sin(math.radians(angle))
        return VectorSpace(x, y)



"""
two_pop_model.py
Models two classes of agents: one tries to follow the other,
the other tries to avoid the first.
"""

# pylint: disable=E731

import logging
import indra.display_methods as disp
import indra.menu as menu
import indra.grid_env as ge
import indra.vs_agent as va
import indra.vector_space as vs

# We represent one stance by a vector pointing along the x-axis, the other
#  by one pointing along the y-axis:
INIT_FLWR = vs.VectorSpace.X_PRE
INIT_LEDR = vs.VectorSpace.Y_PRE

STANCE_TRACKED = INIT_FLWR
STANCE_TINDEX = 0  # the index of the tracked stance in an array of stances


class TwoPopAgent(va.VSAgent):
    """
    An agent taking a stance depending on others' stance.
    variability: controls how self-directed the agent is. The higher the
        variability, the more the agent is influenced by others.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, max_detect=max_move)

        self.my_filter = lambda n: isinstance(n, self.other)

        self.new_stance = vs.VectorSpace()
        self.other = None
        self.variability = variability

    def respond_to_cond(self, env_vars=None):
        """
        Over-riding our parent method to do nothing.
        """
        pass

    def postact(self):
        """
        After we are done acting, adopt our new stance.
        If the stance changes our direction, we adopt the extreme
        of the new direction.
        Then move to an empty cell.
        """
        new_direct = self.new_stance.direction()
        curr_direct = self.stance.direction()
        logging.info("For %s: stance = %s, new stance = %s"
                     % (self.name, str(self.stance), str(self.new_stance)))
        if not new_direct.equals(curr_direct):
            self.direction_changed(curr_direct, new_direct)
            # if adopting a new stance direction, we go to the extreme
            self.new_stance = new_direct
        else:
            self.new_stance = self.new_stance.normalize()
        self.stance = self.new_stance
        self.move_to_empty(grid_view=self.my_view)

    def direction_changed(self, curr_direct, new_direct):
        """
        Some models may need to do something when the direction of an
        agent changes. But in the base class, we don't.
        """
        pass

    def public_stance(self):
        """
        My stance as seen from outside. Filters out internal conflict.
        """
        return self.stance.direction()


class Follower(TwoPopAgent):
    """
    A trend follower: tries to switch to leader's stance
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = Leader
        self.stance = INIT_FLWR

    def eval_env(self, other_pre):
        """
        See how we respond to the stance scene.
        """
        self.new_stance = self.stance.prehend(
            other_pre.intensify(self.variability))
        return self.new_stance


class Leader(TwoPopAgent):
    """
    A leader: avoids follower's stance.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = Follower
        self.stance = INIT_LEDR

    def eval_env(self, other_pre):
        """
        See how we respond to the stance scene.
        For a leader, we reverse what the followers are doing.
        """
        self.new_stance = self.stance.prehend(other_pre.reverse())
        return self.new_stance


class TwoPopEnv(ge.GridEnv):
    """
    A society of leaders and followers.
    Most of the code here is bookkeeping: setting up a census,
        adding menu items, drawing a graph.
    """
    def __init__(self, name, length, height, model_nm=None, torus=False,
                 postact=True, props=None):
        super().__init__(name, length, height, model_nm=model_nm,
                         torus=False, postact=postact, props=props)
        # sub-models will override these vague names with something
        # meaningful in those models
        self.stances = ["yes", "no"]
        self.line_graph_title = "StanceAgents in %s adopting stance %s"
        self.menu.view.add_menu_item("v",
                                     menu.MenuLeaf("(v)iew populations",
                                                   self.view_pop))

    def census(self, disp=True):
        """
        Take a census of our pops.
        Return the total adopting STANCE_TRACKED.
        """
        total_w_stance = 0
        self.user.tell("Populations in period " + str(self.period) +
                       " adopting " +
                       self.stances[STANCE_TINDEX] + ":")
        for var in self.varieties_iter():
            pop = 0
            for agent in self.get_agents_of_var(var):
                if agent.public_stance().equals(STANCE_TRACKED):
                    pop += 1
            total_w_stance += pop
            self.user.tell(var + ": " + str(pop))
            self.append_pop_hist(var, pop)
        return total_w_stance

    def view_pop(self):
        """
        Draw a graph of our changing pops.
        """
        if self.period < 4:
            self.user.tell("Too little data to display")
            return

        (period, data) = self.line_data()
        self.line_graph = disp.LineGraph(self.line_graph_title
                                         % (self.name,
                                            self.stances[STANCE_TINDEX]),
                                         data, period)



"""
grid_env.py

This is an adaptation of the:

Mesa Space Module
=================================

From the GMU Mesa project.

Objects used to add a spatial component to a model.

GridEnv: base grid, a simple list-of-lists.

"""

# Instruction for PyLint to suppress variable name errors,
# since we have a good reason to use one-character variable names for x and y.
# pylint: disable=invalid-name

import math
import random
import itertools
import logging
import indra.node as node
import indra.spatial_env as se

RANDOM = -1

X = 0
Y = 1


def out_of_bounds(x, y, x1, y1, x2, y2):
    """
    Is point x, y off the grid defined by x1, y1, x2, y2?
    """
    return(x < x1 or x >= x2
           or y < y1 or y >= y2)


class Cell(node.Node):
    """
    Cells hold the grid contents.
    They also have a record of where they are in the grid.
    """
    def __init__(self, coords, contents=None):
        super().__init__(None)
        self.__contents = None
        self.coords = coords
        self.contents = contents

    @property
    def contents(self):
        return self.__contents

    @contents.setter
    def contents(self, item):
        old_item = self.__contents
        self.__contents = item
        if item is not None:
            item.cell = self
        if old_item is not None:
            old_item.cell = None

    def is_empty(self):
        """
        Return True if cell empty, else False.
        """
        return not self.contents

    def add_item(self, new_item):
        """
        Add new_item to cell contents.
        Every cell item must have a cell
        field to store its location.
        """
        self.contents = new_item

    def remove_item(self, item):
        """
        If item is our object, set contents to None.
        If that is not our object, do nothing.
        """
        if item == self.contents:
            self.contents = None


class OutOfBounds(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class GridView():
    """
    Defines a subsection of the entire grid env.
    """

    class CellIter:
        """
        Iterate through the view's cells.
        """
        def __init__(self, view):
            self.view = view
            self.x = view.x1
            self.y = view.y1

        def __iter__(self):
            return self

        def __next__(self):
            while self.y < self.view.y2:
                while self.x < self.view.x2:
                    ret = self.view.grid[self.y][self.x]
                    self.x += 1
                    return ret
                self.x = self.view.x1
                self.y += 1

            raise StopIteration()

    def __init__(self, grid, x1, y1, x2, y2):
        """
        see if view is in grid
        adjust x, y to fit if not
        """
        if x1 < 0:
            x1 = 0
        if y1 < 0:
            y1 = 0
        if x2 > grid.width:
            x2 = grid.width
        if y2 > grid.height:
            y2 = grid.height
        self.grid = grid
        self.x1 = x1
        self.x2 = x2
        self.y1 = y1
        self.y2 = y2

    def __iter__(self):
        """
        Iterate through the cells in this view.
        """
        return GridView.CellIter(self)

    def out_of_bounds(self, x, y):
        """
        Returns False if x, y in view, else True.
        """
        return out_of_bounds(x, y, self.x1, self.y1, self.x2, self.y2)

    def filter_view(self, lam):
        """
        Returns a view filtered with lambda functionlam.
        """
        return list(filter(lam, iter(self)))

    def get_empties(self):
        """
        Return all of the unoccupied cells in this view.
        """
        return self.filter_view(lambda x: x.is_empty())

    def get_neighbors(self):
        """
        Return all of the occupied cells in this view.
        """
        return self.filter_view(lambda x: not x.is_empty())


class CompositeView(GridView):
    """
    A composite view combines several other views.
    """
    def __init__(self, grid, views):
        self.grid = grid
        self.views = list(views)  # the parameter will be a tuple

    def __iter__(self):
        """
        Iterate over all our cells: note,
        right now, this return the center cell twice.
        """
        return itertools.chain(self.views)

    def out_of_bounds(self, x, y):
        """
        Is x, y not in this view?
        It is not only if it is not in ANY of the sub-views.
        """
        for view in self.views:
            if not view.out_of_bounds(x, y):
                return False
        return True

    def get_neighbors(self):
        """
        Return all of the occupied cells in this view.
        """
        neighbors = []
        for view in self.views:
            neighbors += view.get_neighbors()
        return neighbors


class GridEnv(se.SpatialEnv):
    """
    Base class for a rectangular grid.

    If a grid is toroidal, the top
    and bottom, and left and right, edges wrap to each other

    Properties:
        width, height: The grid's width and height.
        torus: Boolean which determines whether
            to treat the grid as a torus.

        grid: Internal list-of-lists which holds
            the grid cells themselves.

    Methods:
        get_neighbors: Returns the objects surrounding a given cell.
    """

    def __init__(self, name, width, height, torus=False,
                preact=False, postact=False, model_nm=None, props=None):
        """
        Create a new grid.

        Args:
            height, width: The height and width of the grid
            torus: Boolean whether the grid wraps or not.
        """
        super().__init__(name, width, height,
                         preact=preact, postact=postact,
                         model_nm=model_nm, props=props)

        self.torus = torus
        self.num_cells = width * height

        self.grid = []
        self.empties = []
        for y in range(self.height):
            row = []
            for x in range(self.width):
                cell = self.__new_cell__((x, y))
                row.append(cell)
                self.empties.append(cell)
            self.grid.append(row)

    def __iter__(self):
        # create an iterator that chains the
        #  rows of grid together as if one list:
        return itertools.chain(*self.grid)

    def __getitem__(self, index):
        return self.grid[index]

    def __new_cell__(self, coords):
        return Cell(coords)

    def add_agent(self, agent, position=True):
        """
        Add an agent and link to cell if present in agent.
        """
        super().add_agent(agent, position)
        if agent.cell is not None:
            if agent.cell.contents is not agent:
                agent.cell.add_item(agent)

    def remove_agent(self, agent):
        """
        Remove agent from pop and from grid.
        """
        super().remove_agent(agent)
        if agent.cell is not None:
            agent.cell.remove_item(agent)

    def torus_adj(self, coord, dim_len):
        """
        Convert coordinate, handling torus looping.
        """
        if self.torus:
            coord %= dim_len
        return coord

    def out_of_bounds(self, x, y):
        """
        Is point x, y off the grid?
        """
        return out_of_bounds(x, y, 0, 0, self.width, self.height)

    def get_max_dist(self):
        """
            Args: none

            Returns: The furthest move possible in this env.
        """
        return math.sqrt(self.width**2 + self.height**2)

    def get_col_view(self, col, low=None, high=None):
        """
        Return a view of a single column.
        It will be the whole column from the grid,
        unless low or high are passed.
        """
        if low is None:
            low = 0
        if high is None:
            high = self.height
        return GridView(self, col, low, col + 1, high)

    def get_row_view(self, row, left=None, right=None):
        """
        Return a view of a single row
        It will be the whole row from the grid,
        unless left or right are passed.
        """
        if left is None:
            left = 0
        if right is None:
            right = self.width
        return GridView(self, left, row, right, row + 1)

    def _adjust_coords(self, center, distance, max_val):
        coord1 = max(0, center - distance)
        coord2 = min(max_val, center + distance + 1)
        return (coord1, coord2)

    def get_vonneumann_view(self, center, distance):
        """
        Return a Von Neumann view (row and col)
        centered on center.
        """
        (x, y) = center
        (x1, x2) = self._adjust_coords(x, distance, self.width)
        (y1, y2) = self._adjust_coords(y, distance, self.height)
        col_view = self.get_col_view(x, y1, y2)
        row_view = self.get_row_view(y, x1, x2)
        return CompositeView(self, (col_view, row_view))

    def get_square_view(self, center, distance):
        """
        Attempt to return a view of a square centered on center.
        This might return a non-square rectangle
        if the center is near an edge.
        """
        (center_x, center_y) = center
        (x1, x2) = self._adjust_coords(center_x, distance, self.width)
        (y1, y2) = self._adjust_coords(center_y, distance, self.height)
        return GridView(self, x1, y1, x2, y2)

    def neighbor_iter(self, x, y, distance=1, moore=True, view=None):
        """
        Iterate over our neighbors.
        """
        neighbors = self.get_neighbors(x, y, distance, moore, view)
        return map(lambda x: x.contents, iter(neighbors))

    def get_neighbors(self, x, y, distance=1, moore=True, view=None):
        """
        Return a list of neighbors within a certain purview.

        Args:
            x, y: Coordinates for the neighborhood to get.
            distance: distance, in cells, of neighborhood to get.
            view: we may already have a view we are working with.

        Returns:
            A list of non-None objects in the given neighborhood;
            at most 9 if Moore, 5 if Von-Neumann
            (8 and 4 if not including the center).
        """
        if view is None:
            if moore:
                view = self.get_square_view((x, y), distance)
            else:
                view = self.get_vonneumann_view((x, y), distance)
        return view.get_neighbors()

    def exists_empty_cells(self, grid_view=None):
        """
        Return True if any cells empty else False.
        """
        if len(self.empties) <= 0:
            # if no empties anywhere then none in any view!
            return False
        elif grid_view is not None:
            if len(grid_view.empties) <= 0:
                return False
        return True

    def move_to_empty(self, agent, grid_view=None):
        """
        Moves agent to an empty cell, vacating agent's old cell.
        """
        empty_cell = self.find_empty(grid_view)
        if empty_cell is None:
            logging.error("Agent could not move because no cells are empty")
        else:
            self._move_item(agent, empty_cell)

    def find_empty(self, grid_view=None):
        """
        Return a random, empty cell.
        """
        if self.exists_empty_cells():
            if grid_view is None:
                return random.choice(self.empties)
            else:
                view_empties = grid_view.get_empties()
                if view_empties:
                    return random.choice(view_empties)
        return None

    def position_item(self, item, x=RANDOM, y=RANDOM, grid_view=None):
        """
        Position an agent on the grid.
        This is used when first placing agents! Use 'move_to_empty()'
        when you want agents to jump to an empty cell.
        If x or y are positive, they are used, but if RANDOM,
        we get a random position.
        Ensure this random position is not occupied (in Grid).
        """
        if x == RANDOM or y == RANDOM:
            cell = self.find_empty(grid_view)
            if cell is None:
                return None
        else:
            cell = self._get_cell(x, y)
        self._place_item(cell, item)
        return cell

    def _place_item(self, cell, item):
        """
        Place an agent in the grid.
        """
        cell.add_item(item)
        if cell in self.empties:
            self.empties.remove(cell)

    def _get_contents(self, x, y):
        """
        Extract contents from cell at x, y
        """
        return self._get_cell(x, y).contents

    def move(self, item, x, y):
        """
        Move item from its old cell to cell at x, y.
        """
        dest = self._get_cell(x, y)
        self._move_item(item, dest)

    def _move_item(self, item, dest):
        old_cell = item.cell
        if old_cell is not None:
            old_cell.remove_item(item)
            self._check_empty(old_cell)
        self._place_item(dest, item)

    def _check_empty(self, cell):
        if cell.is_empty():
            self.empties.append(cell)

    def _get_cell(self, x, y):
        return self.grid[y][x]

    def is_cell_empty(self, x, y):
        """
        Returns True if cell is empty, else False.
        A non-existent cell is NOT empty, i.e., not free
        to move to!
        """
        if self.out_of_bounds(x, y):
            return False
        else:
            return self._get_contents(x, y) is None
        
    def dist(self, agent1, agent2):
        """
        Arguments:
            Two grid agents.

        Returns:
            The Euclidian distance between the two
        agents. There isn't numerical ill-conditioning
        with this because the positions are integers. If 
        there are any applications where positions are given
        by nonintegral values, use caution.
        """
        return math.sqrt((agent1.pos[X]-agent2.pos[X])**2 + (agent1.pos[Y]-agent2.pos[Y])**2)

    def free_spot_near(self, agent):
        """
        Looks for an empty cell near agent. If the grid is full,
        returns None. 

        Argument:
            The agent whose nearest empty cell we look for.

        Returns:
            Either (a) the position of this empty cell or
            (b) null in which case the entire grid is full
            of agents.
        """

        max_poss_dist = max(self.width, self.height)
        for i in range(max_poss_dist):
            view = self.get_square_view(center=agent.pos, distance=i)
            for cell in view:
                if cell.is_empty():
                    return cell.coords

        return None



"""
markov.py
The way agents interact is through prehensions.
This implements prehensions as markov chains.
"""

import math
import numpy as np
import random
import indra.prehension as pre
# import logging

ROWS = 0
COLS = 1


def create_iden_matrix(n):
    """
    Create a dim1 * dim2 identity matrix.

    Returns: the new matrix.
    """
    matrix_init = [[] for i in range(n)]
    for i in range(0, n): 
        for j in range(0, n): 
            if i == j:
                matrix_init[i].append(1)
            else:
                matrix_init[i].append(0)

    return np.matrix(matrix_init)


def state_vector(vlen, init_state):
    vals = ""
    for i in range(vlen):
        if i == init_state:
            vals = vals + "1 "
        else:
            vals = vals + "0 "
    return np.matrix(vals)


def probvec_to_state(pv):
    state_vec = None
    cum_prob = 0.0
    r = random.random()
    l = pv.shape[COLS]
    for i in range(l):
        cum_prob += pv.item(i)
        if cum_prob >= r:
            state_vec = state_vector(l, i)
            break
    return state_vec

def get_state(sv):
    """
    This method takes a state vector, sv.
    It then returns the index of the element that is "on,"
    or None if none are.
    """
    i = 0
    for s in sv.flat:
        if s == 1:
            return i
        i += 1
    return None



def from_matrix(m):
    """
    Takes an numpy matrix and returns a prehension.
    """
    pre = MarkovPre("")
    pre.matrix = m
    return pre


def to_matrix(pre):
    return pre.matrix


class MarkovPre(pre.Prehension):
    """
    This class manages taking a state vector and a transition matrix
    and turning them into a new state vector, as well as
    creating matrices more easily than numpy.
    """

    def __init__(self, str_matrix):
        """
        str_matrix is the matrix represented as a string in numpy fashion.
        """
        super().__init__()
        self.matrix = np.matrix(str_matrix)

    def __str__(self):
        return str(self.matrix)

    def prehend(self, other):
        """
        In this class, a prehension prehends another prehension
        as a markov chain process.
        other: prehension to prehend
        """
        if self.matrix.shape[ROWS] == 1:
            return from_matrix(self.matrix * other.matrix)
        else:
            return from_matrix(other.matrix * self.matrix)



"""
utils.py
Various helpful bits that don't fit elsewhere!
"""

import sys
import logging
import indra.prop_args as props

# some values useful for checking valid ranges:
BIG_INT = sys.maxsize
BTWN_ZERO_ONE = (.01, .99)
NTRL_NUMS = (0, BIG_INT)
POS_INTS = (1, BIG_INT)
MAX_GRID = 100000
GRID_LIMITS = (1, MAX_GRID)  # this gives us a max grid of 1 billion elements!
MAX_AGENTS = MAX_GRID * MAX_GRID  # enough to fill a maximum sized grid
AGENT_LIMITS = (1, MAX_AGENTS)


def gen_file_names(model_nm):
    """
    Generate our standard list of I/O spots.
    """
    prog_file = model_nm + ".py"
    log_file = model_nm + ".log"
    prop_file = model_nm + ".props"
    rsul_file = model_nm + ".out"
    return (prog_file, log_file, prop_file, rsul_file)


def run_model(env, prog_file, results_file):
    # Logging is automatically set up for the modeler:
    logging.info("Starting program " + prog_file)

    # And now we set things running!
    env.run()
    env.record_results(results_file)


def read_props(model_nm):
    """
    A prop file to read must be our first arg, if it exists.
    """
    if len(sys.argv) > 1:
        poss_props = sys.argv[1]
        if not poss_props.startswith('-'):  # not a property but a prop file
            return props.PropArgs.read_props(model_nm, poss_props)

    return None


def get_grid_dims(pa, def_dims):
    pa.ask("grid_width", "What is the grid width?", int, default=def_dims,
           limits=GRID_LIMITS)
    pa.ask("grid_height", "What is the grid height?", int, default=def_dims,
           limits=GRID_LIMITS)


def get_agent_num(pa, prop_nm, agent_type, def_num):
    pa.ask(prop_nm, "What is the number of %s?" % (agent_type),
           int, default=def_num, limits=AGENT_LIMITS)


def get_max_move(pa, prop_nm, agent_type, def_move):
    pa.ask(prop_nm, "What is %s's maximum move?" % (agent_type),
           int, default=def_move, limits=AGENT_LIMITS)


def get_pct(pa, prop_nm, agent_type, param_descr, def_pct):
    pa.ask(prop_nm, "What is %s's %s?" % (agent_type, param_descr),
           float, default=def_pct, limits=BTWN_ZERO_ONE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 11 21:13:15 2014
@author: Brandon and Gene
Implements Paul Krugman's babysitting co-op model.
"""

import indra.utils as utils
import indra.prop_args as props
import models.coop_model as cm

MODEL_NM = "coop_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_agent_num(pa, "num_agents", "agents", 100)
        pa.ask("min_holdings",
               "What is a co-op member's minimum desired holding of coupons?",
               float, default=7.5, limits=(1.0, 20.0))
    
    env = cm.CoopEnv(model_nm=MODEL_NM, props=pa)
    
    for i in range(pa.get("num_agents")):
        env.add_agent(
            cm.CoopAgent('agent' + str(i), 5, 0))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
spatial_env.py
The base environment for envs that incorporate space.
"""

from abc import abstractmethod
import logging
import indra.menu as menu
import indra.env as env
import indra.user as user
import indra.display_methods as disp

X = 0
Y = 1


class SpatialEnv(env.Environment):
    """
    Extends the base Environment with entities located in the
    complex plane.
    """
    def __init__(self, name, width, height, preact=True,
                 postact=False, model_nm=None, props=None):

        super().__init__(name, preact=preact,
                         postact=postact, model_nm=model_nm,
                         props=props)

        self.disp_census = True
        self.width = width
        self.height = height
        self.max_dist = self.width * self.height
        self.scatter_plot = None
        self.plot_title = "Agent Positions"
# it only makes sense to plot agents in a spatial env, so add this here:
        self.menu.view.add_menu_item("s",
                                     menu.MenuLeaf("(s)catter plot",
                                                   self.plot))

    def add_agent(self, agent, position=True):
        """
        Add a spatial agent to env
        """
        super().add_agent(agent)
        if position:
            self.position_item(agent)

        logging.debug("Adding " + agent.__str__()
                      + " of variety " + agent.get_type())

    @abstractmethod
    def position_item(self, agent):
        """
        This must be implemented by descendents.
        """

    def closest_x(self, seeker, prehensions):
        """
        What is the closest entity of target_type?
        To be implemented by descendents.
        """
        pass

    def plot(self):
        """
        Show where agents are in graphical form.
        """
        if not disp.plt_present:
            self.user.tell("No graphing package installed", type=user.ERROR)
            return

        data = self.plot_data()
        self.scatter_plot = disp.ScatterPlot(
            self.plot_title, data,
            int(self.width), int(self.height),
            anim=True, data_func=self.plot_data)
        self.scatter_plot.show()

    def plot_data(self):
        data = {}
        for var in self.agents.varieties_iter():
            data[var] = {}
            data[var][X] = []
            data[var][Y] = []
            data[var]["color"] = self.agents.get_var_color(var)
            for agent in self.agents.variety_iter(var):
                if agent.pos is not None:
                    (x, y) = agent.pos
                    data[var][X].append(x)
                    data[var][Y].append(y)
        return data



"""
indra

The indra module of Indra's Net contains the following sub-modules:

    *   agent_pop.py: code dealing with managing and tracking the population of agents.
    *   data_methods.py: getting data about a run out to other programs.
    *   display_methods.py: code dealing with graphs.
    *   entity.py: a mode that describes and environment, agent or object.
    *   entry_point.py: common code for entry points to modules.
    *   env.py: the base environment all others inherit from.
    *   grid_agent.py: the base class for all agents located on a grid.
    *   grid_env.py: the base env for all grid envs.
    *   main_menu.py: the default main menu for all models.
    *   markov.py: basic code for Markov chains.
    *   markov_agent.py: the base class for all Markov agents.
    *   markov_env.py: the base class for Markov environments.
    *   menu.py: contains the basic menuing capabilities.
    *   node.py: the absolute base of everything in Indra's Net.
    *   prop_args.py: the class managing parameters to a model.
    *   spatial_agent.py: the base class for all agents located in space.
    *   spatial_env.py: the base env for spatial models (including grid).
    *   two_pop_model.py: two population models (wolf-sheep etc.) subclass from here.
    *   user.py: manages interacting with the user.
    *   utils.py: miscellaneous useful functions.
    *   vector_space.py: code dealing with vector spaces.
    *   vs_agent.py: base class for agents inhabiting a vector space.

"""



"""
sand_model.py
Abelian sand that has avalanches in interesting patterns.
"""
from math import floor
import indra.display_methods as disp
import indra.grid_agent as ga
import indra.grid_env as ge

X = 0
Y = 1

MAX_SAND = 4
FULL_NEIGHBORHOOD = 4


class SandAgent(ga.GridAgent):
    """
    An agent that collects sand particles until its stack collapses
    onto its neighbors.
    """
    def __init__(self, name, goal, cell):
        super().__init__(name, goal)
        self.cell = cell
        self.sand_amt = 0
        self.ntype = str(self.sand_amt)  # we get our color from our sand amount

    def act(self):
        """
        Our main action goes here, in this case,
        tumbling sand onto our neighbors.
        """
        if self.sand_amt >= MAX_SAND:
            i = 0
            for neighbor in self.neighbor_iter(moore=False, save_hood=True):
                i += 1
                if self.sand_amt > 0:
                    neighbor.add_grains(1)
                    self.sand_amt -= 1
            if self.sand_amt > 0 and i < FULL_NEIGHBORHOOD:
                for j in range(i, FULL_NEIGHBORHOOD):
                    """
                    If at edge, sand "tumbles off."
                    """
                    if self.sand_amt > 0:
                        self.sand_amt -= 1

    def postact(self):
        """
        Here we are going to set our type.
        """
        new_type = str(self.sand_amt)
        if self.ntype != new_type:
            old_type = self.ntype
            self.set_type(new_type)
            self.env.change_agent_type(self, old_type, new_type)

    def add_grains(self, amt):
        """
        Get some sand.
        """
        self.sand_amt += amt


class SandEnv(ge.GridEnv):
    """
    An environment for spilling sand all over the place.
    """

    def __init__(self, name, width, height, model_nm=None, props=None):
        super().__init__(name, width, height, torus=False,
                         model_nm=model_nm, postact=True, props=props)

        self.center_agent = None
        self.set_var_color('0', disp.BLACK)
        self.set_var_color('1', disp.MAGENTA)
        self.set_var_color('2', disp.BLUE)
        self.set_var_color('3', disp.CYAN)
        self.set_var_color('4', disp.RED)
        self.set_var_color('5', disp.YELLOW)
        self.set_var_color('6', disp.GREEN)
        center_x = floor(self.width // 2)
        center_y = floor(self.height // 2)
        print("center = %i, %i" % (center_x, center_y))

        for cell in self:
            (x, y) = cell.coords
            agent = SandAgent("Grainy", "Hold sand", cell)
            self.add_agent(agent, position=False)
            if x == center_x and y == center_y:
                self.center_agent = agent

    def step(self):
        super().step()
        self.center_agent.add_grains(1)



#!/Users/gcallah/miniconda3/bin/python
"""
This is a simple test script. It can be cloned to
create new run scripts, and should be run to test
the system after library changes.
"""

import indra.prop_args as props
import indra.utils as utils
import models.basic_model as bm

# set up some file names:
MODEL_NM = "basic_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a
    # "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_agent_num(pa, "num_agents", "agents", 16)
    
    # Now we create a minimal environment for our agents to act within:
    env = bm.BasicEnv(model_nm=MODEL_NM, props=pa)
    
    # Now we loop creating multiple agents
    #  with numbered names based on the loop variable:
    for i in range(pa.get("num_agents")):
        env.add_agent(bm.BasicAgent(name="agent" + str(i),
                                    goal="acting up!"))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
fashion_model.py
A fashion model that includes followers and hipsters
changing fashions based on each other's choices.
"""
# import logging
import indra.display_methods as disp
import indra.menu as menu
import indra.two_pop_model as tp


class Follower(tp.Follower):
    """
    A fashion follower: tries to switch to hipsters' fashions.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = Hipster


class Hipster(tp.Leader):
    """
    A fashion hipster: tries to not look like followers.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = Follower


class Society(tp.TwoPopEnv):
    """
    A society of hipsters and followers.
    """
    def __init__(self, name, length, height, model_nm=None, torus=False,
                props=None):
        super().__init__(name, length, height, model_nm=model_nm,
                         torus=False, postact=True, props=props)
        self.stances = ["blue", "red"]
        self.line_graph_title = \
            "A. Smith's fashion model: Populations in %s adopting fashion %s"
        self.set_var_color('Hipster', disp.GREEN)
        self.set_var_color('Follower', disp.MAGENTA)
        self.menu.view.add_menu_item("v",
                                     menu.MenuLeaf("(v)iew fashions",
                                                   self.view_pop))



"""
menu_model.py
Model for testing the menu system
"""
import indra.entity as ent
import indra.env as env


class MenuAgent(ent.Agent):
    """
    An agent that just prints that it is around when asked to act
    """

    def __init__(self, name, goal):
        """
        A simple menu init.
        """
        super().__init__(name, goal)

    def act(self):
        """
        Just print my name and goal!
        """
        print("Agent " + self.name + " with a goal of " + self.goal)


class MenuEnv(env.Environment):
    """
    This environment is to test menu features
    """

    def __init__(self, model_nm=None, props=None):
        super().__init__("Menu test",
                         model_nm=model_nm,
                         postact=True,
                         props=props)



"""
menu.py
A menuing class.
The intention is that eventually this menu can be text or GUI based,
depending upon the user's environment.
It allows the addition of submenus at any depth available
memory will allow.
"""

# import logging
from collections import OrderedDict
import indra.node as node
import indra.user as user

MENU_INDENT = 4


class Menu(node.Node):
    """
    Menu items off the main menu or a sub-menu.
    """

    def __init__(self, name, user, level=0):
        super().__init__(name)
        self.user = user
        self.menu_items = OrderedDict()
        self.def_act = None
        self.level = level

    def add_menu_item(self, letter, item, default=False):
        """
        Add an item to this menu.
        """
        self.menu_items[letter] = item
        if default:
            self.def_act = item.act

    def del_menu_item(self, letter):
        """
        Delete an item from this menu.
        """
        if letter in self.menu_items:
            del self.menu_items[letter]

    def act(self):
        """
        What to do when menu is selected.
        """
        return self.display()

    def display(self):
        """
        Display the menu.
        """
        disp_text = ""
        for _ltr, item in self.menu_items.items():
            disp_text += str(item) + " | "
        disp_text = disp_text.rstrip(" | ")
        self.user.tell(disp_text, type=user.MENU,
                           indnt=MENU_INDENT * self.level)

        choice = self.user.ask_for_ltr(
            "Choose one of the above and press Enter: ")
        if choice in self.menu_items:
            self.menu_items[choice].act()
        elif self.def_act is not None:
            self.def_act()
        else:
            pass


class MenuLeaf(node.Node):
    """
    A leaf on the menu tree.
    """

    def __init__(self, name, func):
        super().__init__(name)
        self.func = func

    def act(self):
        return self.func()



"""
auditorium_model.py
Modeling Schelling's auditorium puzzle.
"""
import indra.grid_agent as ga
import indra.grid_env as ge


class AuditoriumAgent(ga.GridAgent):
    """
    Agents that seat themselves in an auditorium.
    """
    def __init__(self, name, goal="Get a seat"):
        super().__init__(name, goal=goal)
        self.seated = False

    def act(self):
        """
        Find a seat.
        Once seated, these agents just stay put!
        """
        if not self.seated:
            self.env.move_to_empty(self)
            self.seated = True

    def is_row_half_empty(self, row):
        num_empties = len(row.get_empties())
        num_neighbors = len(row.get_neighbors())
        return num_empties > num_neighbors


class RearAgent(AuditoriumAgent):
    """
    Agents that want to sit at the back.
    They also want their row less than half full.
    """

    def act(self):
        """
        Find a seat in the furthest row back < 1/2 full.
        Once seated, these agents just stay put!
        """
        if not self.seated:
            for row in self.env.row_iter():
                if self.is_row_half_empty(row):
                    self.env.position_item(self, grid_view=row)
                    self.seated = True
                    break


class NotInFrontAgent(AuditoriumAgent):
    """
    These agents don't want to be the very first person in front.
    They start out in the middle, and fill from there.
    """
    def act(self):
        if not self.seated:
            start = self.env.get_mdl_row()
            for row in self.env.row_iter(start_row=start, direction=-1):
                if self.is_row_half_empty(row):
                    self.env.position_item(self, grid_view=row)
                    self.seated = True
                    break


class Auditorium(ge.GridEnv):
    """
    The auditorium where agents will seat themselves.
    """

    class RowIter:
        """
        Iterate through auditorium's rows.
        Returns a GridView of just that row.
        """
        def __init__(self, aud, start_row=0, direction=1):
            self.auditorium = aud
            self.row = start_row
            self.dir = direction

        def __iter__(self):
            return self

        def __next__(self):
            if self.row < self.auditorium.height:
                ret = self.auditorium.get_row_view(self.row)
                self.row += self.dir
                return ret
            else:
                raise StopIteration()

    def __init__(self, name, height=36, width=40, torus=False,
                 model_nm="Auditorium", num_agents=800, props=None):
        super().__init__(name, width, height, torus=torus,
                         model_nm=model_nm, props=props)
        self.total_agents = num_agents
        self.curr_agents = 0

    def step(self):
        """
        Add an agent each turn.
        """
        super().step()
        if self.curr_agents < self.total_agents:
            self.add_agent(NotInFrontAgent("NotInFront agent %i"
                                           % self.curr_agents),
                           position=False)
            self.curr_agents += 1

    def row_iter(self, start_row=0, direction=1):
        return Auditorium.RowIter(self, start_row, direction)

    def get_mdl_row(self):
        return self.height // 2



"""
user.py
Manages the user for the Indra system.
"""

# import logging
# we are going to do some shenanigans so we can use clint if present
#  and work around if not
MENU = "menu"
PROMPT = "prompt"
ERROR = "error"
INFO = "info"
text_colors = None
clint_present = True
try:
    from clint.textui import colored, puts, indent
    text_colors = {MENU: colored.blue,
                   PROMPT: colored.magenta,
                   ERROR: colored.red,
                   INFO: colored.black}
except ImportError:
    clint_present = False
    text_colors = {MENU: None,
                   PROMPT: None,
                   ERROR: None,
                   INFO: None}

import indra.entity as ent

# user types
TERMINAL = "terminal"
IPYTHON = "iPython"
IPYTHON_NB = "iPython Notebook"


def ask(msg):
    if clint_present:
        puts(text_colors[PROMPT](msg), newline=False)
    else:
        print(msg, end='')
    return input()

def tell(msg, type=INFO, indnt=0):
    if indnt <= 0:
        if clint_present:
            puts(text_colors[type](msg))
        else:
            print(msg)
    else:
        if clint_present:
            with indent(indnt):
                puts(text_colors[type](msg))
        else:
            for i in range(0, indnt):
                msg = '  ' + msg
            print(msg)

class User(ent.Entity):
    """
    We will represent the user to the system as another entity.
    """
    def __init__(self, nm, utype=TERMINAL):
        super().__init__(nm)
        self.utype = utype

    def tell(self, msg, type=INFO, indnt=0):
        """
        Screen the details of output from models.
        """
        if msg and self.utype in [TERMINAL, IPYTHON, IPYTHON_NB]:
            return tell(msg, type=type, indnt=indnt)

    def ask_for_ltr(self, msg):
        """
        Screen the details of input from models.
        """
        choice = self.ask(msg)
        return choice.strip()

    def ask(self, msg):
        """
        Screen the details of input from models.
        """
        assert self.utype in [TERMINAL, IPYTHON, IPYTHON_NB]
        if(self.utype in [TERMINAL, IPYTHON, IPYTHON_NB]):
            return ask(msg)



"""
prehension_agent.py
An agent that use Prehensions to understand its environment.
"""
import indra.grid_agent as ga
import indra.prehension as pre


class PrehensionAgent(ga.GridAgent):
    """
    An agent that use Prehensions to understand its environment.
    """
    def __init__(self, name, goal, max_move=1, max_detect=1):
        super().__init__(name, goal, max_move, max_detect)
        self.my_filter = None
        self.stance = pre.Prehension()

    def survey_env(self):
        """
        Look around and see what surrounds us.
        """
        super().survey_env()
        return None

    def visible_stance(self):
        """
        By default, just our stance. But we may override this.
        For instance, perhaps just our direction is visible to others.
        """
        return self.stance

    def debug_info(self):
        """
        Relevant debugging info.
        """
        s = super().debug_info() + "\nstance: " + str(self.stance)
        return s



from indra.entry_point import entry_point
import models.fashion_run
import models.forestfire_run
import models.sand_run
import models.wolfsheep_run

models = {
             1: ("Adam Smith's Fashion Model", models.fashion_run.run),
             2: ("Forest Fire", models.forestfire_run.run),
             3: ("Abelian Sandpile", models.sand_run.run),
             4: ("Wolf Sheep", models.wolfsheep_run.run)
         }


def main(args=None):
    entry_point(args, models)

if __name__ == "__main__":
    main()



"""
fmarket_model.py
A financial market model that includes chart followers
and value investors.
"""
import logging
import indra.menu as menu
import indra.display_methods as disp
import indra.two_pop_model as tp

INIT_PRICE = 10.0
INIT_ENDOW = INIT_PRICE * 10.0
BUY = tp.INIT_FLWR
SELL = tp.INIT_LEDR


class FinancialAgent(tp.TwoPopAgent):
    """
    A financial agent who trades an asset.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.profit = 0.0
        self.funds = INIT_ENDOW

    def direction_changed(self, curr_direct, new_direct):
        if not new_direct.equals(curr_direct):
            logging.info("For type %s, new_direct = %s, curr_direct = %s"
                         % (type(self).__name__, str(new_direct),
                            str(curr_direct)))
            if new_direct.equals(BUY):
                self.funds -= self.env.asset_price
            else:
                self.funds += self.env.asset_price
            self.calc_profit()

    def calc_profit(self):
        self.profit = self.funds - INIT_ENDOW
        logging.debug("for a = %s, profit (%f) = funds (%f) - INIT_ENDOW (%f)"
                      % (self.name, self.profit, self.funds, INIT_ENDOW))
        self.env.record_profit(self, self.profit)

    def add_env(self, env):
        super().add_env(env)
        if self.stance == BUY:
            self.funds -= self.env.asset_price
            self.calc_profit()


class ChartFollower(FinancialAgent, tp.Follower):
    """
    A trend follower: buys what others are buying.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = ValueInvestor


class ValueInvestor(FinancialAgent, tp.Leader):
    """
    A value investor: buys what others are not buying.
    """
    def __init__(self, name, goal, max_move, variability=.5):
        super().__init__(name, goal, max_move, variability)
        self.other = ChartFollower


class FinMarket(tp.TwoPopEnv):
    """
    A society of value investors and chart followers.
    """
    def __init__(self, name, length, height, model_nm=None, torus=False,
                props=None):
        super().__init__(name, length, height, model_nm=model_nm,
                         torus=False, postact=True, props=props)
        self.total_pop = 0  # to be set once we add agents
        self.asset_price = INIT_PRICE  # an arbitrary starting point
        self.price_hist = []
        self.max_abs_pmove = .5
        self.stances = ["buy", "sell"]
        self.menu.view.add_menu_item("v",
                                     menu.MenuLeaf("(v)iew asset price",
                                                   self.view_price))
        self.follower_profit = 0.0
        self.value_profit = 0.0

    def record_profit(self, agent, profit):
        if isinstance(agent, ChartFollower):
            self.follower_profit += profit
        else:
            self.value_profit += profit

    def census(self, disp=True):
        """
        Take a census of our pops.
        Here we add to our parent an adjustment of the asset price.
        """
        total_buyers = super().census(disp)
        self.user.tell("asset price = %f" % (self.asset_price))
        if self.total_pop == 0:
            # we should only need to do this once, as the model doesn't add
            # agents mid-stream.
            self.total_pop = self.get_total_pop()
        self.adj_asset_price(total_buyers)
        self.price_hist.append(self.asset_price)
        self.user.tell("The chart followers have a net profit of %f."
                       % (self.follower_profit))
        self.user.tell("The value investors have a net profit of %f."
                       % (self.value_profit))
        # these profits are reported anew every exchange, so we have to
        # zero them out after reporting
        self.follower_profit = 0.0
        self.value_profit = 0.0

    def adj_asset_price(self, total_buyers):
        buy_ratio = total_buyers / self.total_pop
        move = self.max_abs_pmove * ((2.0 * buy_ratio) - 1.0)
        if move < 0:  # limit downward moves
            max_move = self.asset_price * .10
            if abs(move) > max_move:
                move = -max_move
        self.asset_price += move

    def view_price(self):
        """
        Draw a graph of our changing asset price.
        """
        if self.period < 4:
            self.user.tell("Too little data to display")
            return

        # put our data in right form for line graph
        data = disp.assemble_lgraph_data("asset price", self.price_hist,
                                         disp.MAGENTA)

        self.line_graph = disp.LineGraph("Asset price history", data,
                                         self.period)



#!/usr/bin/env python3
"""
A predator-prey model with wolves and sheep.
"""

import indra.utils as utils
import indra.prop_args as props
import models.wolfsheep_model as wsm

# set up some file names:
MODEL_NM = "wolfsheep_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 16)
        utils.get_agent_num(pa, "num_wolves", "wolves", 5)
        utils.get_agent_num(pa, "num_sheep", "sheep", 16)
        pa.ask("wolf_repro", "What is the wolf reproduction age?", int, default=16)
        pa.ask("wolf_lforce", "What is the wolf life force?", int, default=7)
        pa.ask("sheep_repro", "What is the sheep reproduction age?", int,
               default=3)
        pa.ask("sheep_lforce", "What is the sheep life force?", int, default=6)
    
    # Now we create a meadow for our agents to act within:
    env = wsm.Meadow("Meadow",
                     pa.get("grid_width"),
                     pa.get("grid_height"),
                     model_nm=MODEL_NM,
                     preact=True,
                     postact=True,
                     props=pa)
    
    # Now we loop creating multiple agents with numbered names
    # based on the number of agents of that type to create:
    for i in range(pa.get("num_wolves")):
        env.add_agent(wsm.Wolf("wolf" + str(i), "Eating sheep",
                               pa.get("wolf_repro"),
                               pa.get("wolf_lforce"),
                               rand_age=True))
    for i in range(pa.get("num_sheep")):
        env.add_agent(wsm.Sheep("sheep" + str(i), "Reproducing",
                                pa.get("sheep_repro"),
                                pa.get("sheep_lforce"),
                                rand_age=True))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
markov_env.py

An environment for Markov chain interactions.

"""

# pylint: disable=invalid-name

import indra.grid_env as ge
import indra.markov as markov

X = 0
Y = 1


class MarkovCell(ge.Cell):
    """
    A grid cell that also has a transition matrix
    specific to it.

    Attribute:
        trans_matrix: a transition matrix
    """
    def __init__(self, coords, contents=None, trans_matrix=None):
        super().__init__(coords, contents)
        self.trans_matrix = trans_matrix

    def __str__(self):
        return str(self.trans_matrix)


class MarkovEnv(ge.GridEnv):
    """
    An env that holds transition matrix for each cell, though strictly speaking,
    it isn't required that the transition matricies must be aquired through
    accessing this cell.
    """

    def __init__(self, name, width, height, trans_str=None, torus=False,
                 matrix_dim=2, model_nm=None, preact=False, postact=False,
                 mobile_agents=False, props=None):
        """
        Create a new markov env. By default the transition matrix is identity.
        """
        if trans_str is None:
            self.def_trans_matrix = markov.from_matrix(markov.create_iden_matrix(matrix_dim))
        else:
            self.def_trans_matrix = markov.MarkovPre(trans_str)
        super().__init__(name, width, height, torus, preact,
                         postact, model_nm=model_nm, props=props)

    def __new_cell__(self, coords):
        return MarkovCell(coords, trans_matrix=self.def_trans_matrix)

    def get_pre(self, agent, n_census):
        cell = self._get_cell(agent.pos[X], agent.pos[Y])
        return cell.trans_matrix

    def set_trans(self, coords, trans):
        cell = self._get_cell(coords[X], coords[Y])
        cell.trans_matrix = trans

    def neighborhood_census(self, agent):
        """
        Counts number of each type in a neighborhood. Stores
        the result in a dictionary.
        """
        n_census = {}

        for neighbor in agent.neighbor_iter():
            if neighbor.ntype in n_census:
                n_census[neighbor.ntype] += 1
            else:
                n_census[neighbor.ntype] = 1

        return n_census



"""
Various models

This module contains miscellaneous example models built using Indra.

The module contains:
    * __main__.py: an entry point into the module.
    * basic_model.py: just exercises the most basic agent capabilities.
    * basic_run.py: runs the above model.
    * coop_model.py: Krugman's babysitting co-op model.
    * coop_run.py: runs the above model.
    * fashion_model.py: Adam Smith's fashion model.
    * fashion_run.py: runs the above model.
    * fmarket_model.py: financial market oscillations.
    * fmarket_run.py: runs the above model.
    * forestfire_model.py: the classic forest fire model.
    * forestfire_run.py: runs the above model.
    * grid_model.py: just exercises the most basic grid capabilites.
    * grid_run.py: runs the above model.
    * menu_model.py: for testing menu capabilities.
    * menu_run.py: runs the above model.
    * obst_model.py: detecting obstacles in the env.
    * obst_run.py: runs the above model.
    * sand_model.py: an Abelian sandpile model.
    * sand_run.py: runs the above model.
    * spatial_model.py: the most basic spatial model.
    * spatial_run.py: runs the above model.
    * timer_model.py: for timing models.
    * timer_run.py: runs the above model.
    * two_pop_model.py: a base for other two population models, like wolfsheep.
    * two_pop_run.py: runs the above model.
    * wolfsheep_model.py: an elementary predator-prey model.
    * wolfsheep_run.py: runs the above model.
"""



"""
markov_agent.py
An agent that use markov chain space Prehensions to understand its environment.
"""
import indra.grid_agent as ga
import indra.markov as markov
import indra.markov_env as markov_env
import indra.prehension_agent as pa


class MarkovAgent(pa.PrehensionAgent):
    """
    An agent whose changeable state determines its action, type, or character.
    Being "Markov-like" an agent will have its next state determined by a 
    probablility distribution represented by a vector. This vector's value is dependent
    only on the agent's current state.

    Attributes:
        state_pre: a state vector prehension
        state: whatever state the agent is in, subject to change
    """
    def __init__(self, name, goal, vlen, init_state, max_detect=1):
        super().__init__(name, goal, max_move=1, max_detect=max_detect)
        self.state_pre = markov.from_matrix(markov.state_vector(vlen,
                                                                init_state))
        self.state = markov.get_state(markov.to_matrix(self.state_pre))

    def survey_env(self):
        """
        Look around and see what surrounds us.
        """
        n_census = self.env.neighborhood_census(self)  # Method found in markov_env.
        return self.env.get_pre(self, n_census)        # Also in markov_env.

    def eval_env(self, other_pre):
        """
        Use the results of surveying the env to decide what to do.
        """
        # print("self_pre before = " + str(self.state_pre))
        prob_pre = self.state_pre.prehend(other_pre)
        self.state_pre.matrix = markov.probvec_to_state(prob_pre.matrix)
        # print("self_pre after = " + str(self.state_pre))
        self.next_state = markov.get_state(self.state_pre.matrix)



#!/usr/bin/env python3
"""
Runs a fashion model with hipsters and followers.
"""

import indra.utils as utils
import indra.prop_args as props
import models.fashion_model as fm

# set up some file names:
MODEL_NM = "fashion_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 16)
        utils.get_agent_num(pa, "num_followers", "followers", 48)
        utils.get_agent_num(pa, "num_hipsters", "hipsters", 16)
        utils.get_max_move(pa, "fmax_move", "follower", 4)
        utils.get_max_move(pa, "hmax_move", "hipster", 4)
        utils.get_pct(pa, "variability", "agent", "variability", .15)
    
    # Now we create a minimal environment for our agents to act within:
    env = fm.Society("Society",
                     pa.get("grid_height"),
                     pa.get("grid_width"),
                     torus=False,
                     model_nm=MODEL_NM,
                     props=pa)
    
    # Now we loop creating multiple agents with numbered names
    # based on the loop variable:
    for i in range(pa.get("num_followers")):
        env.add_agent(fm.Follower("follower" + str(i), 
                                  "Looking like hipsters",
                                  pa.get("fmax_move"), 
                                  pa.get("variability")))
    for i in range(pa.get("num_hipsters")):
        env.add_agent(fm.Hipster("hipster" + str(i), "Looking trendy",
                                 pa.get("hmax_move"), 
                                 pa.get("variability")))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



#!/usr/bin/env python3
"""
This file runs the forestfire_model.
"""

import indra.prop_args as props
import indra.utils as utils
import models.forestfire_model as fm

DEF_GRID_DIM = 80
DEF_DENS = .43
DEF_REGEN = 20
DEF_LIGHTNING = 4

# set up some file names:
MODEL_NM = "forestfire_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a
    # "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, DEF_GRID_DIM)
        utils.get_pct(pa, "density", "forest", "density", DEF_DENS)
        pa.ask("strike_freq", "How many turns between lightning strikes?",
               int, default=DEF_LIGHTNING, limits=utils.POS_INTS)
        pa.ask("regen_period",
               "How many turns before a new tree grows where one has burned?",
               int, default=DEF_REGEN, limits=utils.POS_INTS)
    
    density = pa.get("density")
    grid_x = pa.get("grid_width")
    grid_y = pa.get("grid_height")
    
    # Now we create a forest environment for our agents to act within:
    env = fm.ForestEnv(grid_x, grid_y, density, pa.get("strike_freq"),
                       pa.get("regen_period"),
                       model_nm=MODEL_NM, torus=False,
                       props=pa)
    num_agents = int(grid_x * grid_y * density)
    
    for i in range(num_agents):
        env.add_agent(fm.Tree(name="tree" + str(i)))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
Filename: display_methods.py
Author: Gene Callahan
A collection of convenience functions
for using matplotlib.
"""

from math import ceil
import numpy as np
import networkx as nx
import logging
plt_present = True
try:
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation
    plt.ion()
except ImportError:
    plt_present = False


BLUE = 'b'
RED = 'r'
GREEN = 'g'
YELLOW = 'y'
MAGENTA = 'm'
CYAN = 'c'
BLACK = 'k'
colors = [BLUE, RED, GREEN, YELLOW, MAGENTA, CYAN, BLACK]
NUM_COLORS = len(colors)
X = 0
Y = 1


def hierarchy_pos(graph, root, width=1., vert_gap=0.2, vert_loc=0,
                  xcenter=0.5, pos=None, parent=None):
        """
        This is an attempt to get a tree graph from networkx.
        If there is a cycle that is reachable from root, then this will
        infinitely recurse.
        graph: the graph
        root: the root node of current branch
        width: horizontal space allocated for this branch
                - avoids overlap with other branches
        vert_gap: gap between levels of hierarchy
        vert_loc: vertical location of root
        xcenter: horizontal location of root
        pos: a dict saying where all nodes go if they have been assigned
        parent: parent of this branch.
        """
        if pos is None:
            pos = {root: (xcenter, vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        neighbors = graph.neighbors(root)
        if parent is not None:
            neighbors.remove(parent)
        n_len = len(neighbors)
        dx = 0
        if n_len != 0:
            dx = width / n_len
            nextx = xcenter - width / 2 + dx / 2
            for neighbor in neighbors:
                pos = hierarchy_pos(graph, neighbor, width=dx,
                                    vert_gap=vert_gap,
                                    vert_loc=vert_loc-vert_gap,
                                    xcenter=nextx, pos=pos, parent=root)
                nextx += dx
        return pos


def draw_graph(graph, title, hierarchy=False, root=None):
    """
    Drawing networkx graphs.
    graph is the graph to draw.
    hierarchy is whether we should draw it as a tree.
    """
    if plt_present:
        pos = None
        plt.title(title)
        if hierarchy:
            pos = hierarchy_pos(graph, root)
        nx.draw(graph, pos=pos, with_labels=True)
        plt.show()


def get_color(var, i):
    color = None
    if "color" in var:
        color = var["color"]
    if color is None:
        color = colors[i % NUM_COLORS]
    return color


def assemble_lgraph_data(key, values, color, data=None):
    # put our data in right form for line graph
    if data is None:
        data = {}

    data[key] = {}
    data[key]["data"] = values
    data[key]["color"] = color
    return data


class LineGraph():
    """
    We create a class here to save state for animation.
    Display a simple matplotlib line graph.
    The data is a dictionary with the label
    as the key and a list of numbers as the
    thing to graph.
    data_points is the length of the x-axis.
    """

    def __init__(self, title, varieties, data_points,
                 anim=False, data_func=None):
        self.title = title
        self.anim = anim
        self.data_func = data_func
        self.draw_graph(data_points, varieties)
        if anim:
            animation.FuncAnimation(self.fig,
                                    self.update_plot,
                                    frames=1000,
                                    interval=500,
                                    blit=False)

    def draw_graph(self, data_points, varieties):
        """
        Draw all elements of the graph.
        """
        self.fig, self.ax = plt.subplots()
        x = np.arange(0, data_points)
        self.create_lines(x, self.ax, varieties)
        self.ax.legend()
        self.ax.set_title(self.title)

    def create_lines(self, x, ax, varieties):
        """
        Draw just the data portion.
        """
        self.lines = []
        for i, var in enumerate(varieties):
            data = varieties[var]["data"]
            color = get_color(varieties[var], i)
            y = np.array(data)
            ax.plot(x, y, linewidth=2, label=var, alpha=1.0, c=color)

    def show(self):
        """
        Display the plot.
        """
        plt.show()

    def update_plot(self, i):
        """
        This is our animation function.
        For line graphs, redraw the whole thing.
        """
        plt.clf()
        (data_points, varieties) = self.data_func()
        self.draw_graph(data_points, varieties)
        self.show()


class ScatterPlot():
    """
    We are going to use a class here to save state for our animation
    """

    def __init__(self, title, varieties, width, height,
                 anim=True, data_func=None):
        """
        Setup a scatter plot.
        varieties contains the different types of
        entities to show in the plot, which
        will get assigned different colors
        """
        self.scats = None
        self.anim = anim
        self.data_func = data_func
        self.s = ceil(4096 / width)

        fig, ax = plt.subplots()
        ax.set_xlim(0, width)
        ax.set_ylim(0, height)
        self.create_scats(varieties)
        ax.legend()
        ax.set_title(title)
        plt.grid(True)

        if anim:
            print("Animation is on.")
            animation.FuncAnimation(fig,
                                    self.update_plot,
                                    frames=1000,
                                    interval=500,
                                    blit=False)

    def show(self):
        """
        Display the plot.
        """
        plt.show()

    def update_plot(self, i):
        """
        This is our animation function.
        """
        if self.scats is not None:
            for scat in self.scats:
                if scat is not None:
                    scat.remove()
        varieties = self.data_func()
        self.create_scats(varieties)
        return self.scats

    def get_arrays(self, varieties, var):
        x_array = np.array(varieties[var][X])
        y_array = np.array(varieties[var][Y])
        return (x_array, y_array)

    def create_scats(self, varieties):
        self.scats = []
        for i, var in enumerate(varieties):
            (x_array, y_array) = self.get_arrays(varieties, var)
            if len(x_array) <= 0:  # no data to graph!
                next
            elif len(x_array) != len(y_array):
                logging.debug("Array length mismatch in scatter plot")
                next
            color = get_color(varieties[var], i)
            scat = plt.scatter(x_array, y_array,
                               c=color, label=var,
                               alpha=1.0, marker="8",
                               edgecolors='none', s=self.s)
            self.scats.append(scat)



"""
env.py
Our base environment class for hosting agents.
Other specialized environments inherit from this.
"""

from collections import deque
# import sys
import time
import pdb
import getpass
import networkx as nx
import indra.display_methods as disp
import indra.node as node
import indra.main_menu as mm
import indra.prop_args as pa
import indra.agent_pop as ap
import indra.user as user


class Quit(Exception):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return repr(self.value)


class Environment(node.Node):
    """
    A basic environment allowing starting, stopping,
    stepping, inspection and editing of key objects,  etc.
    """

    prev_period = 0  # in case we need to restore state

    def __init__(self, name, preact=False,
                 postact=False, model_nm=None, props=None):
        super().__init__(name)
        self.graph = nx.Graph()
        pop_name = ""
        if model_nm:
            pop_name += model_nm + " "
        pop_name += "Agents"
        self.agents = ap.AgentPop(pop_name)
        self.graph.add_edge(self, self.agents)
        self.womb = []
        self.period = 0
        self.preact = preact
        self.postact = postact
        self.disp_census = False
        self.line_graph = None
        self.line_graph_title = "Populations in "
        self.model_nm = model_nm
        if props is None:
            if model_nm is not None:
                self.props = pa.PropArgs.get_props(model_nm)
            else:
                self.props = None
        else:
            self.props = props

        user_nm = getpass.getuser()
        user_type = user.TERMINAL
        if self.props is not None:
            self.props.set("user_name", user_nm)
            user_type = self.props.get("user_type", user.TERMINAL)
            self.graph.add_edge(self, self.props)
        self.user = user.User(user_nm, user_type)
        # we have to wait until user is set to do...
        if self.props is None:
            self.user.tell("Could not get props; model_nm = " + str(model_nm),
                           type=user.ERROR)
        self.graph.add_edge(self, self.user)
        self.menu = mm.MainMenu("Main Menu", self)
        self.graph.add_edge(self, self.menu)
        self.graph.add_edge(self, node.universals)

    def add_variety(self, var):
        self.agents.add_variety(var)

    def add_agent(self, agent):
        """
        Add an agent to pop.
        """
        self.agents.append(agent)
        agent.add_env(self)

    def remove_agent(self, agent):
        """
        Remove agent from pop.
        """
        self.agents.remove(agent)

    def join_agents(self, a1, a2):
        """
        Relate two agents.
        """
        self.agents.join_agents(a1, a2)

    def add_child(self, agent):
        """
        Put a child agent in the womb.
        """
        self.womb.append(agent)
        agent.add_env(self)

    def find_agent(self, name):
        """
        Find an agent by name.
        """
        for agent in self.agents:
            if agent.name == name:
                return agent
        return None

    def get_agents_of_var(self, var):
        """
        Return all agents of type 'var'.
        """
        return self.agents.get_agents_of_var(var)

    def get_randagent_of_var(self, var):
        return self.agents.get_randagent_of_var(var)

    def run(self, resume=False, loops=0):
        """
        This is the main menu loop for all models.

        resume: starts up from a previous period.
        loops: run x loops, perhaps for timing purposes.
        """

        if resume:
            self.period = Environment.prev_period
        else:
            self.period = 0

        if loops > 0:
            while self.period < loops:
                self.step()
            self.user.tell("Ran for " + str(self.period) + " periods.")
        else:
            self.user.tell("Welcome, " + self.user.name)
            self.user.tell("Running in " + self.name)
            msg = self.menu.display()
            while msg is None:
                try:
                    msg = self.menu.display()
                    self.user.tell("\nMain Menu; Period: " + str(self.period))
                except Quit:
                    break

            Environment.prev_period = self.period

            self.user.tell(msg)

    def add_menu_item(self, submenu, letter, text, func):
        """
        This func exists to screen the menu class from outside objects:
        no need for them to know more than the env
        """
        self.menu.add_menu_item(submenu, letter, text, func)

    def debug(self):
        """
        Invoke the python debugger.
        """
        pdb.set_trace()

    def eval_code(self):
        """
        Evaluate a line of code.
        """
        eval(self.user.ask("Type a line of code to run: "))

    def list_agents(self):
        """
        List all agents in env.
        """
        self.user.tell("Active agents in environment:")
        for agent in self.agents:
            self.user.tell(agent.name
                           + " with a goal of "
                           + agent.goal)

    def add(self):
        """
        Add a new agent to the env.
        """
        if self.props is not None:
            exec("import " + self.props.get("model")
                 + " as m")
            constr = self.user.ask("Enter constructor for agent to add: ")
            new_agent = eval("m." + constr)
            self.add_agent(new_agent)
        else:
            self.user.tell("Props missing; can't add agent.", type=user.ERROR)

    def agnt_inspect(self):
        """
        View (and possibly alter) an agent's data.
        """
        name = self.user.ask(
            "Type the name of the agent to inspect: ")
        if not name:
            return

        agent = self.find_agent(name.strip())
        if agent is None:
            self.user.tell("No such agent")
        else:
            self.user.tell(agent.debug_info())
        self.edit_field(agent)

    def env_inspect(self):
        """
        Have a look at (and possibly alter) the environment.
        """
        self.pprint()
        self.edit_field(self)

    def edit_field(self, entity):
        """
        Edit a field in an entity.
        """
        while True:
            y_n = self.user.ask("Change a field's value in "
                                + entity.name
                                + "? (y/n) "
                                + "(Only str, float, and int supported.)")
            if y_n in ["Y", "y"]:
                field = self.user.ask("Which field? ")
                nval = self.user.ask("Enter new value for "
                                     + field + ": ")
                fld_type = type(entity.__dict__[field])
                if fld_type is int:
                    entity.__dict__[field] = int(nval)
                elif fld_type is float:
                    entity.__dict__[field] = float(nval)
                else:
                    entity.__dict__[field] = nval
            else:
                break

    def n_steps(self):
        """
        Run for n steps.
        """
        steps = int(self.user.ask("Enter number of steps: "))
        target = self.period + steps
        self.user.tell("Running for %i steps; press Ctrl-c to halt!" % steps)
        time.sleep(3)
        try:
            while self.period <= target:
                step_msg = self.step()
                if step_msg is not None:
                    self.user.tell(step_msg)
                    break
        except KeyboardInterrupt:
            pass

    def cont_run(self):
        """
        Run continuously.
        """
        self.user.tell(
            "Running continously; press Ctrl-c to halt!")
        time.sleep(3)
        try:
            while self.keep_running():
                step_msg = self.step()
                if step_msg is not None:
                    self.user.tell(step_msg)
                    break
        except KeyboardInterrupt:
            pass

    def pwrite(self):
        """
        Write out the properties to a file.
        """
        file_nm = self.user.ask("Choose file name: ")
        if len(file_nm) > 0 and self.props is not None:
            self.props.write(file_nm)

    def disp_props(self):
        """
        Display current system properties.
        """
        if self.props is not None:
            self.user.tell(self.props.display())
        else:
            self.user.tell("Props missing; cannot be displayed.")

    def disp_log(self):
        """
        Display last 16 lines of log file.
        """

        MAX_LINES = 16
        logfile = None

        if self.props is not None:
            logfile = self.props.get_logfile()
        else:
            self.user.tell("Props missing; cannot identify log file!",
                          type=user.ERROR)
            return

        if logfile is None:
            self.user.tell("No log file to examine!", type=user.ERROR)
            return

        last_n_lines = deque(maxlen=MAX_LINES)  # for now hard-coded

        with open(logfile, 'rt') as log:
            for line in log:
                last_n_lines.append(line)

        self.user.tell("Displaying the last " + str(MAX_LINES)
                       + " lines of logfile " + logfile)
        for line in last_n_lines:
            self.user.tell(line.strip())

    def step(self):
        """
        Step period-by-period through agent actions.
        """
        self.census(disp=self.disp_census)

        self.period += 1

# agents might be waiting to be born
        if self.womb is not None:
            for agent in self.womb:
                self.add_agent(agent)
            del self.womb[:]

# there might be state-setting to do before acting
        if self.preact:
            self.preact_loop()

# now have everyone act in random order
        self.act_loop()

# there might be state-setting to do after acting
        if self.postact:
            self.postact_loop()

    def act_loop(self):
        """
        Loop through randomly through agents calling their act() func.
        """
        for agent in self.agents.agent_random_iter():
            agent.act()

    def preact_loop(self):
        """
        Loop through agents calling their preact() func.
        """
        for agent in self.agents:
            agent.preact()

    def postact_loop(self):
        """
        Loop through agents calling their postact() func.
        """
        for agent in self.agents.agent_random_iter():
            agent.postact()

    def graph_agents(self):
        """
        Draw a graph of the agent relationships.
        """
        self.agents.draw()

    def graph_env(self):
        """
        Draw a graph of the env's relationships.
        """
        self.draw()

    def graph_unv(self):
        """
        Draw a graph of the universal relationships.
        """
        node.universals.draw()

    def graph_class_tree(self):
        node.Node.class_draw()

    def keep_running(self):
        """
        Placeholder
        """
        return True

    def view_pop(self):
        """
        Graph our population levels.
        """
        if not disp.plt_present:
            self.user.tell("No graphing package installed", type=user.ERROR)
            return
        if self.period < 4:
            self.user.tell("Too little data to display", type=user.ERROR)
            return

        (period, data) = self.line_data()
        self.line_graph = disp.LineGraph(self.line_graph_title + self.name,
                                         data, period)
        self.line_graph.show()

    def line_data(self):
        return (self.period, self.agents.get_pop_hist())

    def plot(self):
        """
        Placeholder
        """
        self.user.tell("Plot not implemented in this model", type=user.ERROR)

    def quit(self):
        """
        Leave this run.
        """
        self.user.tell("Exiting main loop.")
        raise Quit("User exit")

    def contains(self, agent_type):
        """
        Do we have this sort of thing in our env?
        """
        return self.agents.contains(agent_type)

    def census(self, disp=True, exclude_var=None):
        """
        Take a census of what is in the env.
        """
        results = self.agents.census(exclude_var=exclude_var)
        if disp:
            self.user.tell("Populations in period "
                           + str(self.period) + ":")
            self.user.tell(results)

    def varieties_iter(self):
        return self.agents.varieties_iter()

    def variety_iter(self, var):
        return self.agents.variety_iter(var)

    def set_var_color(self, var, color):
        self.agents.set_var_color(var, color)

    def get_pop(self, var):
        """
        Return the population of variety 'var'
        """
        return self.agents.get_pop(var)

    def get_total_pop(self):
        """
        Return the total number of agents.
        """
        return self.agents.get_total_pop()

    def get_my_pop(self, agent):
        """
        Return the population of agent's type
        """
        return self.agents.get_my_pop(agent)

    def change_agent_type(self, agent, old_type, new_type):
        self.agents.change_agent_type(agent, old_type, new_type)

    def record_results(self, file_nm):
        self.agents.record_results(file_nm)

    def change_pop_data(self, var, change):
        self.agents.change_pop_data(var, change)

    def get_pop_data(self, var):
        return self.agents.get_pop_data(var)

    def append_pop_hist(self, var, pop):
        self.agents.append_pop_hist(var, pop)

    def get_var_pop_hist(self, var):
        return self.agents.get_var_pop_hist(var)



"""
Filename: height_model.py
Author: Gene Callahan and Brandon Logan
"""

import indra.entity as ent
import indra.env as env
import random
import indra.display_methods as disp

REPRODUCE = "reproduce"
ENV_NM = "Schelling height model"
MIN_HEIGHT = .2
CHILD_HEIGHT_VAR = 10
RUNT_PCT = .67


class HeightAgent(ent.Agent):
    """
    An agent who reproduces and has offspring with
    a deviation around its height.
    """
    def __init__(self, name, height, parent_height):
        super().__init__(name, REPRODUCE)
        self.height = height
        self.alive = True
        self.mychild = None
        self.parent_height = parent_height

    def act(self):
        self.reproduce()
        self.alive = False

    def get_new_height(self):
        """
        Calculate the height of my child.
        We use parent_height to generate reversion to the mean.
        """
        mu = (self.height + self.parent_height) / 2
        new_height = random.gauss(mu, mu / CHILD_HEIGHT_VAR)
        return new_height

    def reproduce(self):
        """
        Produce offspring.
        """
        new_height = self.get_new_height()
        self.mychild = self.__class__(self.name + str(self.env.period),
                                      new_height, self.height)
        self.env.add_child(self.mychild)


class HeightAgentEng(HeightAgent):

    def reproduce(self):
        """
        Produce offspring that are not under 'runt_height'.
        """
        super().reproduce()
        if self.mychild.height < self.env.runt_height:
            self.mychild.height = self.env.runt_height


class HeightEnv(env.Environment):
    """
    This class creates an environment for Schelling height agents
    """

    def __init__(self, model_nm=None, props=None):
        super().__init__("Height Environment", model_nm=model_nm, preact=True,
                        props=props)
        self.avg_height = {}
        self.runt_height = 0

    def census(self, disp=True):
        """
        Take a census of our pops.
        """
        for var in self.varieties_iter():
            total_height = 0
            num_agents = self.get_pop(var)

            # perhaps we should use reduce here:
            for agent in self.variety_iter(var):
                total_height += agent.height

            self.avg_height[var] = total_height / num_agents
            self.append_pop_hist(var, self.avg_height[var])
            if var == "HeightAgentEng":
                self.runt_height = RUNT_PCT * self.avg_height[var]
            self.user.tell("\nAverage %s height for period %i: %f" %
                           (var, self.period, self.avg_height[var]))

    def view_pop(self):
        """
        Draw a graph of our changing pops.
        """
        if self.period < 4:
            self.user.tell("Too little data to display")
            return

        (period, data) = self.line_data()
        self.line_graph = disp.LineGraph("Schelling's height model",
                                         data, period, anim=False)

    def preact_loop(self):
        for agent in reversed(self.agents):
            if not agent.alive:
                self.remove_agent(agent)



# -*- coding: utf-8 -*-
"""
Created on Mon Mar  2 21:40:05 2015

@authors: Brandon Logan
    Gene Callahan
Implements Thomas Schelling's segregation model.
An agent moves when she finds herself to be "too small"
of a minority in a particular neighborhood.
"""


import random
import indra.vector_space as vs
import indra.vs_agent as va
import indra.grid_env as grid

MOVE = True
STAY = False
RED = vs.X
BLUE = vs.Y
RED_PRE = vs.VectorSpace.X_PRE
BLUE_PRE = vs.VectorSpace.Y_PRE
RED_AGENT = "RedAgent"
BLUE_AGENT = "BlueAgent"
AGENT_TYPES = {RED: RED_AGENT, BLUE: BLUE_AGENT}


class SegregationAgent(va.VSAgent):
    """
    An agent that moves location based on its neighbors' types
    """
    def __init__(self, name, goal, min_tol, max_tol, max_move=100,
                 max_detect=1):
        super().__init__(name, goal, max_move=max_move, max_detect=max_detect)
        self.tolerance = random.uniform(max_tol, min_tol)
        self.stance = None
        self.orientation = None
        self.visible_pre = None

    def eval_env(self, other_pre):
        """
        Use the results of surveying the env to decide what to do.
        """
        # no neighbors, we stay put:
        if other_pre.equals(vs.VectorSpace.NULL_PRE):
            return STAY

        # otherwise, see how we like the hood
        other_pre = other_pre.normalize()
        other_projection = other_pre.project(self.orientation)
        my_projection = self.stance.project(self.orientation)
        if other_projection < my_projection:
            return MOVE
        else:
            return STAY

    def respond_to_cond(self, eval_vars=None):
        """
        If we don't like the neighborhood, we jump to a random empty cell.
        """
        self.move_to_empty()

    def visible_stance(self):
        """
        Our visible stance differs from our internal one.
        It is just our "color."
        """
        return self.visible_pre


class BlueAgent(SegregationAgent):
    """
    We set our stance.
    """
    def __init__(self, name, goal, min_tol, max_tol, max_move=100,
                 max_detect=1):
        super().__init__(name, goal, min_tol, max_tol,
                         max_move=max_move, max_detect=max_detect)
        self.orientation = BLUE
        self.visible_pre = BLUE_PRE
        self.stance = vs.stance_pct_to_pre(self.tolerance, BLUE)


class RedAgent(SegregationAgent):
    """
    We set our stance.
    """
    def __init__(self, name, goal, min_tol, max_tol, max_move=100,
                 max_detect=1):
        super().__init__(name, goal, min_tol, max_tol,
                         max_move=max_move, max_detect=max_detect)
        self.orientation = RED
        self.visible_pre = RED_PRE
        self.stance = vs.stance_pct_to_pre(self.tolerance, RED)


class SegregationEnv(grid.GridEnv):
    """
    The segregation model environment, mostly concerned with bookkeeping.
    """

    def __init__(self, name, width, height, torus=False,
                 model_nm="Segregation", props=None):

        super().__init__(name, width, height, torus=False,
                         model_nm=model_nm, props=props)
        self.plot_title = name
        # setting our colors adds varieties as well!
        self.set_var_color(AGENT_TYPES[BLUE], 'b')
        self.set_var_color(AGENT_TYPES[RED], 'r')
        self.num_moves = 0
        self.move_hist = []
        self.menu.view.del_menu_item("v")  # no line graph in this model

    def move_to_empty(self, agent, grid_view=None):
        super().move_to_empty(agent, grid_view)
        self.num_moves += 1

    def census(self, disp=True):
        """
        Take a census recording the number of moves.
        """
        self.move_hist.append(self.num_moves)
        self.user.tell("Moves per turn: " + str(self.move_hist))
        self.num_moves = 0

    def record_results(self, file_nm):
        """
        """
        f = open(file_nm, 'w')
        for num_moves in self.move_hist:
            f.write(str(num_moves) + '\n')
        f.close()



#!/usr/bin/env python3
"""
A script to test our grid capabilities.
"""

import indra.utils as utils
import indra.prop_args as props
import models.sand_model as sm

# set up some file names:
MODEL_NM = "grid_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 200)
    
    # Now we create a minimal environment for our agents to act within:
    env = sm.SandEnv("Abelian sand env",
                     pa.get("grid_width"),
                     pa.get("grid_height"),
                     model_nm=MODEL_NM,
                     props=pa)
    
    # This env adds agents itself.
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
agent_pop.py

Manages agent information.
The AgentPop class implements a flexible data structure
that allows the user of the class to access agents
from multiple perspectives.
"""

from collections import OrderedDict
import random
import itertools
import json
import networkx as nx
import logging
import indra.node as node
import indra.entity as ent

MAX_ZERO_PER = 8


class AgentPop(node.Node):
    """
    Holds our collection of agents and a sub-graph
    of their relationships.
    This should implement list functions.
    """

    def __init__(self, name):
        super().__init__(name)
        self.vars = OrderedDict()
        self.graph = nx.Graph()

    def __iter__(self):
        alists = []
        for var in self.varieties_iter():
            alists.append(self.vars[var]["agents"])
        # create an iterator that chains the lists together as if one:
        return itertools.chain(*alists)

    def __len__(self):
        l = 0
        for var in self.varieties_iter():
            l += len(self.vars[var]["agents"])
        return l

    def __reversed__(self):
        all_agents = self.all_agents_list()
        return reversed(all_agents)

    def varieties_iter(self):
        """
        Allows iteration over the varieties of agents.
        """
        return iter(self.vars)

    def variety_iter(self, var):
        """
        Allows iteration over the agents of one variety
        """
        return iter(self.vars[var]["agents"])

    def all_agents_list(self):
        all_agents = []
        for var in self.varieties_iter():
            all_agents += self.vars[var]["agents"]
        return all_agents

    def agent_random_iter(self):
        """
        Loop through agents in random order.
        """
        all_agents = self.all_agents_list()
        random.shuffle(all_agents)
        return iter(all_agents)

    def element_at(self, i):
        """
        Another way to treat the AgentPop as if it were really
        one big list.
        """
        if i < 0 or i > len(self):
            raise IndexError()
        else:
            for var in self.varieties_iter():
                l = len(self.vars[var]["agents"])
                logging.debug("Looking for element from "
                              + var + " at position "
                              + str(i) + " and var has len "
                              + str(l))
                if i < l:
                    # that means the agent is in this list
                    return self.vars[var]["agents"][i]
                else:
                    # otherwise, the agent lies in one of the
                    # remaining lists, so subtract the length
                    # of this one from i and continue.
                    i -= l

    def get_var_color(self, var):
        """
        Get the display color for this variety.
        """
        return self.vars[var]["disp_color"]

    def set_var_color(self, var, color):
        """
        Set the display color for this var.
        """
        self.add_variety(var)
        self.vars[var]["disp_color"] = color

    def add_variety(self, var):
        """
        Sometimes, we may know a variety is coming, and want to
        add it even before any agents of that variety are created,
        for census or graphing purposes.
        """
        if var not in self.vars:
            self.vars[var] = {"agents": [],
                              # some arbitrary data to track for pop:
                              "pop_data": 0,
                              "pop_hist": [],
                              "my_periods": 0,
                              "disp_color": None}

    def append(self, agent, v=None):
        """
        Appends to agent list.
        """
        if v is None:
            var = agent.get_type()
        else:
            var = v
        logging.info("Adding %s of variety %s" % (agent.name, var))

        if var not in self.vars:
            self.add_variety(var)
            self.graph.add_edge(self, var)

        self.vars[var]["agents"].append(agent)

# we link each agent to the variety
# so we can show their relationship
        self.graph.add_edge(var, agent)

    def remove(self, agent, v=None):
        """
        Removes from agent list.
        """
        logging.debug("Removing " + agent.name + " from agents")
        if v is None:
            var = agent.get_type()
        else:
            var = v
        self.vars[var]["agents"].remove(agent)
        self.graph.remove_node(agent)  # also removes edges!

    def join_agents(self, a1, a2):
        """
        Add a graph edge between agents.
        """
        self.graph.add_edge(a1, a2)

    def contains(self, var):
        """
        Do we have this sort of thing in our env?
        """
        return var in self.vars

    def get_agents_of_var(self, var):
        """
        Return all agents of type var.
        """
        if var in self.vars:
            return self.vars[var]["agents"]
        else:
            return None

    def get_randagent_of_var(self, var):
        return random.choice(self.vars[var]["agents"])

    def get_pop(self, var):
        """
        Return the population of variety 'var'
        """
        return len(self.vars[var]["agents"])

    def get_total_pop(self):
        total_pop = 0
        for var in self.varieties_iter():
            total_pop += self.get_pop(var)
        return total_pop

    def get_my_pop(self, agent):
        """
        Return the population of agent's type
        """
        var = agent.get_type()
        return self.get_pop(var)

    def get_var_pop_hist(self, var):
        return self.vars[var]["pop_hist"]

    def get_pop_hist(self):
        """
        Make a list containing the population history
        for each var in vars, if var is None.
        We should merge this with display_methods
        assemble data when we can.
        """
        pop_hist = {}
        for var in self.varieties_iter():
            pop_hist[var] = {}
            pop_hist[var]["data"] = self.vars[var]["pop_hist"]
            pop_hist[var]["color"] = self.vars[var]["disp_color"]
            logging.debug("Setting color for %s to %s"
                          % (var, pop_hist[var]["color"]))
        return pop_hist

    def get_pop_data(self, var):
        """
        Return the value of pop_data for 'var'.
        """
        return self.vars[var]["pop_data"]

    def change_pop_data(self, var, change):
        """
        Change the value of pop_data by 'change.'
        """
        self.vars[var]["pop_data"] += change

    def change_agent_type(self, agent, old_type, new_type):
        self.remove(agent, v=old_type)
        self.append(agent, v=new_type)

    def append_pop_hist(self, var, pop):
        """
        Add the most recent pop to pop_hist.
        """
        self.vars[var]["pop_hist"].append(pop)

    def census(self, exclude_var=None):
        """
        Take a census of agents by variety.
        Return a string of results for possible display.
        """
        ret = ""
        for v in self.varieties_iter():
            if v != exclude_var:
                pop = self.get_pop(v)
                ret += v + ": " + str(pop) + "\n"
                var = self.vars[v]
                var["pop_hist"].append(pop)
                # a type might enter the env after period 0!
                # so we track that in my_periods
                var["my_periods"] += 1
        return ret

    def jsondump(self, obj):
        if type(obj) == nx.Graph:
            return "Graph"
        elif isinstance(obj, ent.Entity):
            return obj.to_json()
        else:
            return obj.__dict__

    def record_results(self, file_nm):
        json.dump(self, open(file_nm, 'w'), indent=4,
                  default=self.jsondump)



import sys

import bigbox.__main__
import models.__main__
import schelling.__main__

MODULE_NAME = 0
MODULE_MAIN = 1

modules = {1: ("Big Box", bigbox.__main__.main),
           2: ("General models", models.__main__.main),
           3: ("Schelling models", schelling.__main__.main)
           }

def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    print("The modules containing models are: ")
    for num, module in modules.items():
        print(str(num) + ". " + module[MODULE_NAME])
    choice = int(
                input("Enter module number to view avialable models: "))
    modules[choice][MODULE_MAIN]()

if __name__ == "__main__":
    main()



"""
node.py
A very basic class implementing some thing at the bottom of our
net of objects, such as being a node in a graph.
"""

# import logging
import inspect
import networkx as nx
import indra.display_methods as disp

UNIVERSAL = "universal"

indras_net = nx.Graph()


def get_class_name(genera):
    return genera.__name__


class Node():
    """
    Contains a name and a graph.
    """

    class_graph = nx.Graph()  # we are going to graph our class hierarchy
    node_added = False

    @classmethod
    def class_draw(cls):
        """
        This draws our class hierarchy.
        """
        if not disp.plt_present:
            return

        if cls.class_graph is not None:
            disp.draw_graph(cls.class_graph, "Class Hierarchy",
                            hierarchy=True, root=Node.__name__)

    @classmethod
    def connect_to_class_tree(cls):
        for c in inspect.getmro(cls):
            if c != cls:
                if c.__name__ not in Node.class_graph:
                    c.connect_to_class_tree()
                    break
                else:
                    Node.class_graph.add_edge(get_class_name(c),
                                              get_class_name(cls))
                    return
        cls.connect_to_class_tree()  # call until all upper classes are in

    def __init__(self, name):
        self.name = name
# every node is potentially a graph itself
        self.graph = None
        self.ntype = self.__class__.__name__
        if not Node.node_added:
            Node.class_graph.add_node(Node.__name__)
        if self.ntype not in Node.class_graph:
            self.__class__.connect_to_class_tree()

    def __str__(self):
        return self.name

    def draw(self):
        """
        Every node should have some way to draw itself.
        The default is to draw a graph.
        Other types of nodes will need different draw functions.
        """
        if not disp.plt_present:
            return

        if self.graph is not None:
            disp.draw_graph(self.graph, self.name)

    def display(self):
        """
        Every node should have some way to display itself.
        The difference between draw and display:
        Display makes the object show up on screen.
        Draw reveals the object's structure (such
        as graphing its components).
        """
        pass

    def debug_info(self):
        """
        Relevant debugging info.
        """
        s = "name: " + self.name + "\nntype: " + self.ntype
        return s

    def get_type(self):
        """
        Returns node's type.
        By default, this is its class name.
        """
        return self.ntype

    def set_type(self, new_type):
        """
        Change node's type.
        Use caution: A lot depends on this field!
        In particular, AgentPop uses it to sort agents.
        """
        self.ntype = new_type

    def to_json(self):
        return {"name": self.name}


class Universals(Node):
    """
    Our universals, the things which relate instances or classes.
    This is not implemented at the moment.
    """

    universals_graphed = False

    def __init__(self):
        super().__init__("Universals")
        self.unis = {}
        self.graph = nx.MultiDiGraph()
        if not Universals.universals_graphed:
            pass

    def add_universal(self, uni):
        """
        We want to be able to view all
        univeral relationships at a glance,
        so we will store them here
        """
        self.unis[uni] = []

universals = Universals()



#!/usr/bin/env python3
"""
Set up and run the auditorium model.
"""

import indra.utils as utils
import indra.prop_args as props
import schelling.auditorium_model as am

# set up some file names:
MODEL_NM = "auditorium_model"

def run():
    (prog_file, log_file, 
     prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 8)
        utils.get_agent_num(pa, "num_agents", "agents", 16)
    
    # Now we create a minimal environment for our agents to act within:
    env = am.Auditorium("Auditorium",
                        height=pa.get("grid_height"),
                        width=pa.get("grid_width"),
                        torus=False,
                        model_nm=MODEL_NM,
                        num_agents=pa.get("num_agents"),
                        props=pa)
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
grid_agent.py
Overrides pos for SpatialAgent.
Maybe other things for grid as well.
"""

# import logging
import indra.spatial_agent as sa

X = 0
Y = 1


class GridAgent(sa.SpatialAgent):
    """
    This class is the parent of all entities that are
    located in a grid space (and might or might not move in it)
    """

    def __init__(self, name, goal, max_move=1, max_detect=1, cell=None):
        super().__init__(name, goal, max_move=max_move,
                         max_detect=max_detect)
        self.__cell = cell
        self.neighborhood = None
        self.hood_size = max_detect
        self.my_view = None

    @property
    def cell(self):
        return self.__cell

    @cell.setter
    def cell(self, cell):
        """
        Sets where we are on grid.
        """
        self.__cell = cell

    @property
    def pos(self):
        if self.cell:
            return self.cell.coords
        else:
            return None

    def _neighbor_filter(self, moore=True, view=None, filt_func=None):
        """
        This filter can be overridden in class's that need a different one
        by passing in a different filter function.
        """
        if filt_func is None:
            def filt_func(n): return n is not self
        return filter(filt_func,
                      self.env.neighbor_iter(self.pos[X], self.pos[Y],
                                             moore=moore,
                                             view=view))

    def act(self):
        """
        Here we set up the usual pattern of action for a grid agent.
        """
        env_vars = self.survey_env()
        eval_vars = self.eval_env(env_vars)
        if eval_vars:
            self.respond_to_cond(eval_vars)

    def survey_env(self):
        """
        Have a look around, by default in a square view.
        """
        self.my_view = self.get_square_view(self.hood_size)
        return (self.my_view)

    def eval_env(self, env_vars):
        """
        Return True if we need to do something.
        """
        return env_vars is not None

    def respond_to_cond(self, eval_vars=None):
        """
        Default is to jump to empty cell.
        """
        self.move_to_empty(grid_view=self.my_view)

    def get_square_view(self, distance):
        return self.env.get_square_view(self.pos, distance)

    def neighbor_iter(self, moore=True, save_hood=False, view=None,
                      filt_func=None, sq_v=1):
        """
        Iterate over our neighbors.
        In some models, the neighbors don't move:
            then we can save the neighborhood and a lot of overhead!
        """
        if view is None:
            view = self.my_view
        if view is None:
            # our default view is a square reaching 1 cell out from self
            # in every direction
            view = self.get_square_view(sq_v)

        if not save_hood:
            return self._neighbor_filter(moore, view=view, filt_func=filt_func)
        else:
            if not self.neighborhood:
                self.neighborhood = list(self._neighbor_filter(
                                         moore, view=view,
                                         filt_func=filt_func))
            return iter(self.neighborhood)

    def move_to_empty(self, grid_view=None):
        self.env.move_to_empty(self, grid_view=grid_view)

    def debug_info(self):
        """
        Relevant debugging info.
        """
        s = (super().debug_info() + "\nLocation x: "
             + str(self.pos[X]) + ", y: " + str(self.pos[X])
             + "\nHood size: " + str(self.hood_size))
        return s

    def to_json(self):
        """
        We're going to make a dictionary of the 'safe' parts of the object to
        output to a json file. (We can't output the env, for instance, since
        IT contains a reference to each agent!)
        """
        safe_fields = super().to_json()
        safe_fields["pos"] = self.pos
        return safe_fields



"""
The default main menu for our ABMs.
"""

import networkx as nx
from indra.menu import Menu, MenuLeaf

class MainMenu(Menu):
    """
    Our basic menu.
    """

    def __init__(self, name, env):
        super().__init__(name, env.user)
        self.env = env
        e = self.env  # just a shorthand to make the following lines shorter
        self.graph = nx.Graph()

# file menu
        self.file = Menu("(f)ile", e.user, level=1)
        self.add_menu_item("f", self.file)
        self.file.add_menu_item("e", MenuLeaf("(e)xamine log", e.disp_log))
        self.file.add_menu_item("q", MenuLeaf("(q)uit", e.quit))
        self.file.add_menu_item("w", MenuLeaf("(w)rite props", e.pwrite))

# edit menu
        self.edit = Menu("(e)dit", e.user, level=1)
        self.add_menu_item("e", self.edit)
        self.edit.add_menu_item("a", MenuLeaf("(a)dd agent", e.add))
        self.edit.add_menu_item("i",
                                MenuLeaf("(i)nspect agent",
                                         e.agnt_inspect))
        self.edit.add_menu_item("e",
                                MenuLeaf("inspect (e)nv", e.env_inspect))

# view menu
        self.view = Menu("(v)iew", e.user, level=1)
        self.add_menu_item("v", self.view)
        self.view.add_menu_item("l", MenuLeaf("(l)ist agents",
                                              e.list_agents))
        self.view.add_menu_item("p", MenuLeaf("(p)roperties", e.disp_props))
        self.view.add_menu_item("v", MenuLeaf("(v)iew populations",
                                              e.view_pop))
# graph submenu
        self.graph = Menu("(g)raph", e.user, level=1)
        self.view.add_menu_item("g", self.graph)
        self.graph.add_menu_item("a", MenuLeaf("(a)gents", e.graph_agents))
        self.graph.add_menu_item("c", MenuLeaf("(c)lasses",
                                               e.graph_class_tree))
        self.graph.add_menu_item("e", MenuLeaf("(e)nvironment",
                                               e.graph_env))
        self.graph.add_menu_item("u", MenuLeaf("(u)niversals", e.graph_unv))

# tools menu
        self.tools = Menu("(t)ools", e.user, level=1)
        self.add_menu_item("t", self.tools, default=True)
        self.tools.add_menu_item("s", MenuLeaf("(s)tep", e.step),
                                 default=True)
        self.tools.add_menu_item("r", MenuLeaf("(r)un", e.cont_run))
        self.tools.add_menu_item("n", MenuLeaf("run (n) steps", e.n_steps))
        self.tools.add_menu_item("d", MenuLeaf("(d)ebug", e.debug))



"""
forestfire_model.py
Adapted from the George Mason mesa project.
This model shows the spread of a forest fire through a forest.
"""

import indra.display_methods as disp
import indra.markov as markov
import indra.markov_agent as ma
import indra.markov_env as menv


X = 0
Y = 1

# tree condition strings
HEALTHY = "Healthy"
ON_FIRE = "On Fire"
BURNED_OUT = "Burned Out"
NEW_GROWTH = "New Growth"

# states
HE = 0
OF = 1
BO = 2
NG = 3

STATE_MAP = { HE: HEALTHY, OF: ON_FIRE, BO: BURNED_OUT, NG: NEW_GROWTH }

NSTATES = 4

NORMAL_TRANS = ".9995 .0005 0 0; 0 0 1 0; 0 0 .90 .10; 1 0 0 0"
FIRE_TRANS = "0 1 0 0; 0 0 1 0; 0 0 .90 .10; .30 .70 0 0"

class Tree(ma.MarkovAgent):
    '''
    A tree cell.

    Attributes:
        condition: Can be "New Growth", "Healthy", "On Fire", or "Burned Out"

    '''
    def __init__(self, name):
        '''
        Create a new tree.
        '''
        super().__init__(name, "health", NSTATES, HE)
        self.state = HE
        self.ntype = STATE_MAP[HE]
        self.next_state = None


    def is_healthy(self):
        return self.state == HE or self.state == NG

    def set_state(self, new_state):
        """
        Set tree's new type.
        """
        old_type = self.ntype
        self.state = new_state
        self.ntype = STATE_MAP[new_state]
        self.env.change_agent_type(self, old_type, self.ntype)

    def postact(self):
        """
        Set our type to next_state.
        """
        if self.next_state is not None and self.next_state != self.state:
            # print("Setting state to " + str(self.next_state))
            self.set_state(self.next_state)
            self.next_state = None

        return self.pos


class ForestEnv(menv.MarkovEnv):
    '''
    Simple Forest Fire model.
    '''
    def __init__(self, width, height, density, strike_freq, regen_period,
                 torus=False, model_nm="ForestFire", postact=True,
                 props=None):
        '''
        Create a new forest fire model.

        Args:
            width, height: The size of the grid to model
            density: What fraction of grid cells have a tree in them.
        '''
        # Initialize model parameters
        super().__init__("Forest Fire", width, height, NORMAL_TRANS,
                         torus=False, model_nm=model_nm, postact=postact,
                         props=props)
        self.density = density
        self.plot_title = "A Forest Fire"

        # setting our colors adds varieties as well!
        self.set_var_color(BURNED_OUT, disp.BLACK)
        self.set_var_color(ON_FIRE, disp.RED)
        self.set_var_color(HEALTHY, disp.GREEN)
        self.set_var_color(NEW_GROWTH, disp.CYAN)

        # set up our two possible transition matrices:
        self.normal = markov.MarkovPre(NORMAL_TRANS)
        self.fire = markov.MarkovPre(FIRE_TRANS)

    def spread_fire(self, tree):
        """
        We change the transition matrix of the surrounding cells
        to the fire matrix.
        """
        self.set_trans(tree.pos, self.fire)

    def burn_out(self, tree):
        """
        We change the transition matrix of this tree
        to the normal matrix.
        """
        self.set_trans(tree.pos, self.normal)

    def get_pre(self, agent, n_census):
        if (ON_FIRE in n_census) and (n_census[ON_FIRE] > 0):
            return self.fire
        else:
            return self.normal



#!/usr/bin/env python3
"""
A script to test our two pop capabilities.
"""

import indra.utils as utils
import indra.prop_args as props
import indra.two_pop_model as itpm
import models.two_pop_model as tpm

# set up some file names:
MODEL_NM = "two_pop_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    
    # We store basic parameters in a "property" file; this allows us to save
    #  multiple parameter sets, which is important in simulation work.
    #  We can read these in from file or set them here.
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_grid_dims(pa, 6)
        utils.get_agent_num(pa, "num_agents", "agents", 8)
    
    # Now we create a minimal environment for our agents to act within:
    env = itpm.TwoPopEnv("Test grid env",
                     pa.get("grid_width"),
                     pa.get("grid_height"),
                     torus=False,
                     model_nm=MODEL_NM,
                     postact=True,
                     props=pa)
    
    # Now we loop creating multiple agents with numbered names
    # based on the loop variable:
    
    # Get Number of Followers
    for i in range(pa.get("num_agents")):
        env.add_agent(tpm.TestFollower(name="Follower" + str(i),
                      goal="taking up a grid space!",max_move=1))
    
    # Get Number of Hipsters
    for i in range(pa.get("num_agents")):
        env.add_agent(tpm.TestLeader(name="Leader" + str(i),
                      goal="taking up a grid space!",max_move=1))
    
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
plane_model.py
You can clone this file and its companion plan_run.py
to easily get started on a new spatial model.
It also is a handy tool to have around for testing
new features added to the base system.
"""
import indra.plane_agent as pa
import indra.plane_env as pe


class TestPlaneAgent(pa.MobileAgent):
    """
    An agent that just prints where it is when asked to act
    """

    def act(self):
        print(pe.pos_msg(self, self.pos))



"""
wolfsheep_model.py
Wolves and sheep roaming a meadow, with wolves eating sheep
that get near them.

Available under the terms of the GNU GPL
"""
import random
import indra.markov as markov
import indra.markov_agent as ma
import indra.markov_env as menv
import numpy as np
import operator as op
import math

X = 0
Y = 1

# agent condition strings
NORTH = "North"
SOUTH = "South"
EAST = "East"
WEST = "West"

NSTATES = 4

N = 0
S = 1
E = 2
W = 3

STATE_MAP = { N: NORTH, S: SOUTH, E: EAST, W: WEST }

class Creature(ma.MarkovAgent): 
    """
    A Creature moves around intelligently based on his
    prehension of the env.

    Attributes: 
        alive: boolean indicating if Creature should exist in the env 
        age: int keeps trak of how many steps Creature has been in existance; 
            used for controling periodic reproduction
        other: Wolf or Sheep object to be used by subclass to identify what
            Creature is different than itself
        repro_age: if age (mod repro_age) == 0, make a new Creature of same
            class.
        life_force: if this int goes to or below zero, Creature dies and is 
            removed from env.
        init_life_force: int 
        speed: an int amount of times a Creature may move per turn
        state: the direction the Creature is moving
    """
    def __init__(self, name, goal, repro_age, life_force, init_state, max_detect=1,
                 rand_age=False, speed=1):
        super().__init__(name, goal, NSTATES, init_state, max_detect=max_detect)
        if not rand_age:
            self.age = 0
        else:
            self.age = random.randint(0, repro_age - 2)
        self.alive = True
        self.other = None
        self.repro_age = repro_age
        self.life_force = life_force
        self.init_life_force = life_force
        self.speed = speed
        self.state = init_state

    def died(self):
        """
        Removes dead agent from grid.
        """
        if self.alive:
            self.alive = False
            self.env.died(self)

    def act(self):
        """
        The Creature moves either one unit north, south, east, 
        or west. For however many units of "speed" he has, he 
        can move that many times per action.
        """
        for i in range(self.speed):
            super().act()
            self.state = self.next_state
            self.move(self.state)
            
    def move(self, state):
        """
        Moves self one unit in a cardinal direction if
        that direction is empty and doesn't lead off the
        board.

        Args:
            self: the agent to be moved
            state: a cardinal direction
                North: N = 0
                South: S = 1
                East: E = 2
                West: W = 3
        """
        x = self.pos[X]
        y = self.pos[Y]
        if state == N:
            if self.env.is_cell_empty(x, y+1):
                self.env.move(self, x,y+1)
        elif state == S:
            if self.env.is_cell_empty(x, y-1):
                self.env.move(self, x,y-1)
        elif state == E:
            if self.env.is_cell_empty(x-1, y):
                self.env.move(self, x-1,y)
        else:
            if self.env.is_cell_empty(x+1, y):
                self.env.move(self, x+1,y)
            
    def postact(self):
        """
        Every step ages the Creature by 1. If the Creature 
        has not lifeforce, he dies. If the Creature happens to
        be in the right stage of life, he may reproduce.
        """
        self.age += 1
        self.life_force -= 1
        if self.life_force <= 0:
            self.died()
        elif self.age % self.repro_age == 0:
            self.reproduce()

    def reproduce(self):
        """
        Adds a Creature of self's type to a random place in the env.
        """
        if self.alive:
            creature = self.__class__(self.name + "x", self.goal,
                                      self.repro_age, self.init_life_force)
            self.env.add_agent(creature)


class Wolf(Creature):
    """
    A wolf: moves around randomly and eats any sheep
    nearby.

    Attributes: 
        other: the class of the other kind of Creature, Sheep
        ntype: "node type," variety of Creature,
            used for bookkeeping purposes by agent_pop.
            
    """
    def __init__(self, name, goal, repro_age, life_force, max_detect=10,
                    rand_age=False, speed=2):
        init_state = random.randint(0,3)
        super().__init__(name, goal, repro_age, life_force, init_state,
                            max_detect=max_detect, rand_age=rand_age, speed=speed)
        self.other = Sheep
        self.ntype = "Wolf"

    def preact(self):
        """
        After (or before depending on how you view it) everything moves, 
        wolves eat nearby sheep.
        """
        creatures = self.neighbor_iter()
        for creature in creatures:
            if type(creature) is Sheep:
                self.eat(creature)

    def eat(self, sheep):
        """
        Gains sheep's life force and removes sheep from env.
        """
        self.life_force += sheep.life_force
        sheep.died()


class Sheep(Creature):
    """
    A sheep: moves when wolf is nearby and sometimes gets eaten.

    Attributes: 
        other: the class of the other kind of Creature, Wolf
        ntype: variety of Creature, used for bookkeeping purposes by agent_pop, not 
            an essential feature for the running of this program
    """
    def __init__(self, name, goal, repro_age, life_force, max_detect=5,
                 rand_age=False, speed=1):
        init_state = random.randint(0,3)
        super().__init__(name, goal, repro_age, life_force, init_state,
                         max_detect=max_detect, rand_age=rand_age, speed=speed)
        self.other = Wolf
        self.ntype = "Sheep"


class Meadow(menv.MarkovEnv):
    """
    A meadow in which wolf eat sheep.
    """

    def died(self, prey):
        """
        Removes prey from env.
        """
        self.remove_agent(prey)

    def get_pre(self, agent, n_census):
        """
        Gathers and uses env info to make a prehension transition matrix

        Args: 
            self: the env
            agent: the agent acting in the env 
            n_census: unused
        Returns:
            prehesion based on what Creatures are North, South,
            East, and West of self, determining what direction
            the agent might come.
        """

        trans_str = ""

        d, total = self.dir_info(agent)

        if(type(agent) == Wolf):
            trans_str += self.wolf_trans(d, total)
        else:
            trans_str += self.sheep_trans(d, total)
        
        trans_matrix = markov.from_matrix(np.matrix(trans_str))
        return trans_matrix

    def dir_info(self, agent):
        """
        We count wolves and sheep that are North, South, East, West 
        of the agent in question. What quadrant they're in, and how
        far they are from the agent factors into the information returned.
        -----------------------------------
        | \                             / | Positive Diagonal Line: y = x + (ya-xa)
        |   \                         /   |                         x = y - (ya-xa)
        |     \          N          /     |
        |       \                 /       |
        |         \             /         |
        |           \         /           | 
        |             \     /             |
        |               \ /               |
        |     E   agent @ (xa,ya)   W     |
        |               / \               | 
        |             /     \             |
        |           /         \           |
        |         /             \         |
        |       /                 \       |
        |     /                     \     |
        |   /            S            \   |
        | /  other @ (othr[X],othr[Y])  \ | Negative Diagonal Line: y = -x + (ya+xa)
        -----------------------------------                         x = -y + (ya+xa)

        Args:
            self: the environment 
            agent: the agent surveying its environment 

        Returns:
            A dict maping cardinal directions to numbers representing
            both how far and how many agents of the opposite type are
            in that cardinal direciton.
        """
        directions = {N: 0, S: 0, E: 0, W: 0}
        creatureCoords = (0,0)
        total = 0
        creatures = agent.neighbor_iter(sq_v=10)
        for creature in creatures:
            if type(creature) == agent.other:
                othr = creature.pos
                xa = agent.pos[X]
                ya = agent.pos[Y]
                # North constitutes upper quadrant and positive diagonal line
                if(-othr[X]+(ya+xa) < othr[Y] and othr[Y] >= othr[X] + (ya-xa)):
                    dist = math.sqrt((ya-othr[Y])**2 + (xa-othr[X])**2)
                    directions[N] += 1/dist
                    total += 1/dist
                # South constitutes lower quadrant and negative diagonal line
                elif(othr[X]+(ya-xa) >= othr[Y] and othr[Y] < -othr[X]+(ya+xa)):
                    dist = math.sqrt((ya-othr[Y])**2 + (xa-othr[X])**2)
                    directions[S] += 1/dist
                    total += 1/dist
                # East constitutes left quadrant and negative diagonal line
                elif(othr[Y]-(ya-xa) > othr[X] and othr[X] <= -othr[Y]+(ya+xa)):
                    dist = math.sqrt((ya-othr[Y])**2 + (xa-othr[X])**2)
                    directions[E] += 1/dist
                    total += 1/dist
                # West constitutes right quadrant and positive diagonal line
                else:
                    dist = math.sqrt((ya-othr[Y])**2 + (xa-othr[X])**2)
                    directions[W] += 1/dist
                    total += 1/dist

        return directions, total

    def wolf_trans(self, d, total):
        """
        The wolf uses its survey of the environment to see which
        directions have the closest and most sheep. He'll probably
        go in the direction with the highest reward.

        Args:
            d: a dictionary containing information about where wolves are
             relative to the sheep 
            total: the sum of all the numerical elements of d; used in
             computation

        Returns:
            trans_str: the string representing the sheep's possibilities
            for movement
        """
        trans_str = ""

        NS, NE, NW, NN = 0,0,0,0
        SN, SE, SW, SS = 0,0,0,0
        EN, ES, EW, EE = 0,0,0,0
        WN, WS, WE, WW = 0,0,0,0

        # Random movement if nothing is in view.
        if total == 0:
            NS, NE, NW, NN = 0.25,0.25,0.25,0.25
            SN, SE, SW, SS = 0.25,0.25,0.25,0.25
            EN, ES, EW, EE = 0.25,0.25,0.25,0.25
            WN, WS, WE, WW = 0.25,0.25,0.25,0.25
        else:
            NS = d[S]/total
            NE = d[E]/total
            NW = d[W]/total
            NN = 1 - NS - NE - NW

            SN = d[N]/total
            SE = d[E]/total
            SW = d[W]/total
            SS = 1 - SN - SE - SW 

            EN = d[N]/total
            ES = d[S]/total
            EW = d[W]/total
            EE = 1 - EN - ES - EW

            WN = d[N]/total
            WS = d[S]/total
            WE = d[E]/total
            WW = 1 - WN - WS - WE

        trans_str += str(NN) + " " + str(NS) + " " + str(NE) + " " + str(NW) + ";"
        trans_str += str(SN) + " " + str(SS) + " " + str(SE) + " " + str(SW) + ";"
        trans_str += str(EN) + " " + str(ES) + " " + str(EE) + " " + str(EW) + ";"
        trans_str += str(WN) + " " + str(WS) + " " + str(WE) + " " + str(WW)

        return trans_str

    def sheep_trans(self, d, total):
        """
        The sheep uses it's survey of the environment to determine which
        directions are the least dangerious. He'll go one of these directions.

        Args:
            d: a dictionary containing information about where wolves are
             relative to the sheep 
            total: the sum of all the numerical elements of d; used in
             computation

        Returns:
            trans_str: the string representing the probability a sheep will
                move in any cardinal direction
        """
        best_dir = min(d, key=d.get)

        num_close = 0

        col_N = 0
        col_S = 0
        col_E = 0
        col_W = 0
        if(d[N]==d[best_dir]):
            num_close += 1
            col_N = 1
        if(d[S]==d[best_dir]):
            num_close += 1
            col_S = 1
        if(d[E]==d[best_dir]):
            num_close += 1
            col_E = 1
        if(d[W]==d[best_dir]):
            num_close += 1
            col_W = 1

        NN = col_N/num_close
        SN = col_N/num_close
        EN = col_N/num_close
        WN = col_N/num_close

        NS = col_S/num_close
        SS = col_S/num_close
        ES = col_S/num_close
        WS = col_S/num_close

        NE = col_E/num_close
        SE = col_E/num_close
        EE = col_E/num_close
        WE = col_E/num_close

        NW = col_W/num_close
        SW = col_W/num_close
        EW = col_W/num_close
        WW = col_W/num_close

        trans_str = ""

        trans_str += str(NN) + " " + str(NS) + " " + str(NE) + " " + str(NW) + ";"
        trans_str += str(SN) + " " + str(SS) + " " + str(SE) + " " + str(SW) + ";"
        trans_str += str(EN) + " " + str(ES) + " " + str(EE) + " " + str(EW) + ";"
        trans_str += str(WN) + " " + str(WS) + " " + str(WE) + " " + str(WW)

        return trans_str



"""
Big Box

The bigbox module contains code relating to studying the relationship between
big box retailers and mom-and-pop shops.

The module contains:
    * big_box_model.py: the basic model.
    * big_box_run.py: collects parameters for the big box model and runs it.
"""



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 11 20:23:01 2014

@author: Brandon
This implements Paul Krugman's babysitting co-op model.
"""

# import logging
import indra.entity as ent
import indra.env as env
import indra.display_methods as disp
import random

''' These are our possible goals'''
UNKNOWN = 'unknown'
BABYSIT = 'babysit'
GO_OUT = 'go out'

ENV_NM = "Krugman Babysitting Co-op"


class CoopAgent(ent.Agent):
    """
    An agent that chooses to babysit or go out based on coupon holdings
    """
    def __init__(self, name, coupons, min_holdings):
        super().__init__(name, UNKNOWN)
        self.coupons = coupons
        self.min_holdings = min_holdings
        self.sitting = False

    def act(self):
        self.sitting = False
        if self.coupons < self.min_holdings:
            self.goal = BABYSIT
        elif self.coupons > self.min_holdings:
            if random.random() > .5:
                self.goal = GO_OUT
            else:
                self.goal = BABYSIT

    def pay(self, sitter):
        self.coupons -= 1
        sitter.coupons += 1


class CoopEnv(env.Environment):
    """
    An environment the co-op agents use to survey and interact
    """

    def __init__(self, model_nm=ENV_NM, props=None):
        super().__init__("Co-op Environment",
                         model_nm=model_nm, postact=True,
                         props=props)
        self.rd_exchanges = 0

    def postact_loop(self):
        """
        If an agent wants to go out, it hopes to 
        find a babysitter, and pay them
        """
        exchange = 0
        for agent in reversed(self.agents):
            if agent.goal == GO_OUT:
                sitter = self.assign_sitter()
                if sitter is None:
                    pass
                else:
                    agent.pay(sitter)
                    exchange += 1
            else:
                pass
            self.rd_exchanges = exchange
        print('\nCoupon exchanges in round ' + str(self.period)
              + ': ' + str(self.rd_exchanges) + '\n')

    def assign_sitter(self):
        """
        Surveys the environment for available and willing babysitters
        """
        for agent in reversed(self.agents):
            if agent.goal == BABYSIT:
                if agent.sitting is False:
                    agent.sitting = True
                    return agent
                else:
                    pass
            else:
                pass

    def census(self, disp=True):
        """
        Take a census of our pops.
        """
        self.change_pop_data('CoopAgent', self.rd_exchanges)
        for var in self.varieties_iter():
            pop = self.get_pop_data(var)
            self.append_pop_hist(var, pop)
        self.change_pop_data('CoopAgent', -self.rd_exchanges)

    def view_pop(self):
        """
        Draws a line graph of coupon exchanges
        """
        if self.period < 4:
            print("Too little data to display")
            return

        (period, data) = self.line_data()
        self.line_graph = disp.LineGraph("History of Coupon Exchanges",
                                         data, period)



#!/usr/bin/env python3
"""
A script to test our spatial capabilities.
"""

import indra.utils as utils
import indra.prop_args as props
import indra.spatial_env as se
import models.spatial_model as sm

# set up some file names:
MODEL_NM = "spatial_model"
(prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)

# We store basic parameters in a "property" file; this allows us to save
#  multiple parameter sets, which is important in simulation work.
#  We can read these in from file or set them here.
pa = utils.read_props(MODEL_NM)
if pa is None:
    pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
    pa.set("num_agents", 2)

# Now we create a minimal environment for our agents to act within:
env = se.SpatialEnv("Test spatial env", 100.0, 100.0,
                    model_nm=MODEL_NM, props=pa)

# Now we loop creating multiple agents with numbered names
# based on the loop variable:
for i in range(pa.get("num_agents")):
    env.add_agent(
        sm.TestSpatialAgent(name="agent" + str(i),
                            goal="moving around aimlessly!"))

utils.run_model(env, prog_file, results_file)



"""
obst_model.py
Testing obstacle detection.
"""
# import logging
from collections import deque
import random
import indra.grid_agent as ga

WEST = (-1, 0)
NW = (-1, 1)
NORTH = (0, 1)
NE = (1, 1)
EAST = (1, 0)
SE = (1, -1)
SOUTH = (0, -1)
SW = (-1, -1)
DIRECTIONS = [WEST, NW, NORTH, NE, EAST, SE, SOUTH, SW]


def get_rand_vector_mag(dist):
    vector_mag = random.randint(1, dist)
    return vector_mag


def get_rand_direction():
    return random.choice(DIRECTIONS)


def get_steps(x, y, num_steps, x_step, y_step):
    steps = []
    print("num_steps = %i" % (num_steps))
    for i in range(0, num_steps):
        x += x_step
        y += y_step
        steps.append((x, y))
    print(str(steps))
    return steps


class Obstacle(ga.GridAgent):
    """
    An obstacle.
    """

    def __init__(self, name, goal="Be in the way."):
        super().__init__(name, goal, max_move=0, max_detect=0)

    def act(self):
        pass


class ObstacleAgent(ga.GridAgent):
    """
    An agent that avoids obstacles in its env.
    Properties:
        max_move: the furthest agent can move in one turn
        tolerance: how close we feel OK getting near an obstacle
    """

    def __init__(self, name, goal, max_move=2, max_detect=2,
                 tolerance=1):
        super().__init__(name, goal, max_move=max_move,
                         max_detect=max_detect)
        self.tolerance = tolerance
        self.wbound = None
        self.hbound = None

    def add_env(self, env):
        """
        We capture our env's dimensions so we don't
        have to keep fetching them.
        If we create envs that re-size themselves,
        this code must change!
        """
        super().add_env(env)
        self.wbound = self.env.width - 1
        self.hbound = self.env.height - 1

    def act(self):
        """
        Our act is to move to a random cell,
        on a straight line or a diagonal.
        """
        (x, y) = self.pos
        print("In oa act(); max_move = %i" % self.max_move)
        vector_mag = get_rand_vector_mag(self.max_move)
        print("vector_mag = %i" % (vector_mag))
        (x_step, y_step) = get_rand_direction()
        steps = get_steps(x, y, vector_mag + (self.tolerance - 1),
                          x_step, y_step)
        (new_x, new_y) = self._check_for_obst(steps, vector_mag, x, y)
        print("In oa act(); x = %i, y = %i, new_x = %i, new_y = %i"
              % (x, y, new_x, new_y))
        self.env.move(self, new_x, new_y)

    def _check_for_obst(self, steps, vector_mag, clear_x, clear_y):
        """
        Look for obstacles in the path 'steps'
        and stop short if they are there.
        clear_x and clear_y start out as the agent's current pos,
        which must, of course, be clear for this agent to occupy!
        If there happens to be an obstacle less than tolerance - 1
        from the agent's initial square, well, we just stay put.
        """
        lag = self.tolerance - 1
        lookahead = deque(maxlen=lag + 1)
        for (x, y) in steps:
            lookahead.append((x, y))
            if not self.env.is_cell_empty(x, y):
                return (clear_x, clear_y)
            elif lag > 0:
                lag -= 1
            else:
                (clear_x, clear_y) = lookahead.popleft()
        return (clear_x, clear_y)



"""
Schelling models

This module contains models based on the work of Thomas Schelling.

The module contains:
    * auditorium_model.py: Schelling's auditorium filling puzzle.
    * auditorium_run.py: runs the above model.
    * height_model.py: Schelling's genetically engineered height model.
    * height_run.py: runs the above model.
    * segregation_model.py: Schelling's model of neighborhood segregation.
    * segregation_run.py: runs the above model.
"""



"""
Filename: data_methods.py
Author: Gene Callahan
A collection of functions for exporting Indra data
"""

def pop_report(file_nm, varieties):
    """
    Write CSV file with pop data.
    Args:
        file_nm: file to write
        varieties: data to output
    """
    if len(file_nm) > 0:
        f = open(file_nm, "w")
        cols = []
        head = ''
        for i, var in enumerate(varieties):
            head += (var + ",")
            list = []
            cols.append(
                varieties[var]["data"])
        head = head[:-1]  # remove last comma!
        f.write(head + "\n")
        num_cols = len(cols)
        if num_cols > 0:
            num_rows = len(cols[0])
            for i in range(0, num_rows):
                srow = ''
                for j in range(0, num_cols):
                    srow += (str(cols[j][i]) + ',')
                srow = srow[:-1]  # remove last comma!
                f.write(srow + "\n")
        f.close()




# created based on C. Warrick's page here:
# https://chriswarrick.com/blog/2014/09/15/python-apps-the-right-way-entry_points-and-scrip    ts/

import sys
import os

MODEL_NAME = 0
MODEL_RUN = 1


def entry_point(args, models):
    print("In this module, the models available to run are: ")
    for num, model in models.items():
        print(str(num) + ". " + model[MODEL_NAME])
    choice = int(input("Enter model number to run it: "))
    models[choice][MODEL_RUN]()



"""
Filename: spatial_agent.py
Author: Gene Callahan
"""

from collections import deque
# import logging
import indra.entity as ent

MAX_EXCLUDE = 10


class SpatialAgent(ent.Agent):
    """
    This class is the parent of all entities that are
    located in space (and might or might not move in it)
    """

    def __init__(self, name, goal, max_move=0, max_detect=1):
        super().__init__(name, goal)
        self.max_move = max_move
        self.max_detect = max_detect
        self.__pos = None
        self.exclude = deque(maxlen=MAX_EXCLUDE)

    @property
    def pos(self):
        return self.__pos

    @pos.setter
    def pos(self, pos):
        self.__pos = pos



"""
prehension.py
The way agents interact is through prehensions.
This is an abstract implementation of a prehension.
Sub-class this to instantiate an implementation.
"""

from abc import abstractmethod
# import logging


class Prehension:
    """
    All prehensions must have the following properties:
        The operation +, which we will call “prehend”, accepts two prehensions
            as arguments and produces a third prehension.
        Axioms:
            a.  Closure: Every prehending involving two prehensions will produce
                a prehension.
            b.  Associativity: (a + b) + c = a + (b + c)
            In a typical agent model, this will mean that we must ensure that,
                say, a neighborhood can interact with a neighborhood (b + c),
                and then with an agent (a + (b + c)). Furthermore, this must
                produce an identical prehension to that produced by an agent
                interacting with one neighborhood and then another one
                ((a + b) + c).
            c.  Identity: Any prehension prehending the null prehension produces
                unchanged.
            d.  Invertibility: For any prehension, there is another prehension
                that combines with it to produce the null prehension.
        prehensions MAY also have another operator, *, such that:
        The operation *, which we will call “intensify” (although it may also
            de-intensify) accepts an element of R and an element of G
            (a prehension), and produces an element of G.
        Axioms:
            a.  a, b member G:
                i.  (a + b)x = ax + bx
                ii. a(x + y) = ax + ay

    """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def prehend(self, other):
        """
        A prehension prehends another prehension.
        Must return a prehension.
        """
        return Prehension()


    def equals(self, other):
        """
        However descendants decide if prehensions are equal.
        """
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Sep 18 16:45:50 2014

@author: Gene Callahan and Brandon Logan
"""

import indra.utils as utils
import indra.prop_args as props
import schelling.height_model as hm

START_HEIGHT = 100.0

MODEL_NM = "height_model"

def run():
    (prog_file, log_file, prop_file, results_file) = utils.gen_file_names(MODEL_NM)
    pa = utils.read_props(MODEL_NM)
    if pa is None:
        pa = props.PropArgs(MODEL_NM, logfile=log_file, props=None)
        utils.get_agent_num(pa, "num_agents", "agents", 80)
    env = hm.HeightEnv(model_nm=MODEL_NM, props=pa)
    for i in range(pa.get("num_agents")):
            env.add_agent(
                hm.HeightAgentEng('Eng agent' + str(i),
                                  START_HEIGHT, START_HEIGHT))
            env.add_agent(
                hm.HeightAgent('agent' + str(i), START_HEIGHT, START_HEIGHT))
    utils.run_model(env, prog_file, results_file)

if __name__ == "__main__":
    run()



"""
big_box_model.py
Simulates a small town with consumers.

The consumers may shop at either "Mom and Pop" and
chain "Big Box" stores. The
town's consumers prefer the local stores, but will shot at the 
big box retailer when convenient. The major difference between Mom and 
Pops and Big Boxes is the larger endowment Big Boxes have initially.

Initially there are only local stores, but later the big box stores come in and 
change the population dynamic.
"""

import random
import collections
import math
import numpy as np
import indra.utils as utils
import indra.menu as menu
import indra.grid_agent as ga
import indra.grid_env as ge
import indra.display_methods as disp
import indra.data_methods as data

MODEL_NM = "big_box_model"

pa = utils.read_props(MODEL_NM)

X = 0
Y = 1

NA = -1.0    # good not available

MOM_POP = 0
BIG_BOX = 1
NUM_STATES = 2

BB_MULT = 1000
BB_DIV = 200

MIN_ADJ = 0.01  # never offer 0 utils for a purchase

# types of goods sold
HARDWARE = 0
HAIRCUT = 1
GROCERIES = 2
BOOKS = 3
COFFEE = 4
NUM_GOODS = 5

GOODS_MAP = {0: "Hardware", 1: "Barber", 2: "Grocery",
             3: "Book Store", 4: "Coffee Shop"}


class MarketParticipant(ga.GridAgent):
    """
    All agents in the small town participate in the economy. Only some
    sell goods of any type.
    """

    def __init__(self, name, goal, init_state=0):
        super().__init__(name, goal, NUM_STATES, init_state)

    def sells(self, good):
        """
        Does this participant sell this good?
        """
        return False


class Consumer(MarketParticipant):
    """
    Everyday consumer of EverytownUSA. He has a preference for the cosy
    small shops stores, but will buy occasionally from big boxes.

    Attributes:
        state: the number describing the agent's preference 
        for a type of marketparticipant 
        allowance: The amount the agent will buy from a store.
    """

    def __init__(self, name, goal, init_state, allowance):
        super().__init__(name, goal, init_state)
        self.state = init_state
        self.allowance = allowance
        self.last_utils = 0.0      # what statisfaction did they get from
                                   # previous purchase?
        self.view = None

    def survey_env(self):
        """
        We survey the whole environment to see what 
        stores sell the good we want.

        Args: self
        Returns: 
            store_count: a list of all the marketparticipants in agent's travel
                distance that sell the good he wants.
        """
        if self.view is None:
            self.view = self.env.get_square_view(center=self.pos,
                distance=math.sqrt(self.env.width**2 + self.env.height**2))
        sellers = []
        sellers.extend(self.neighbor_iter(
                       view=self.view,
                       filt_func=lambda x: x.sells(self.goal)))
        return sellers

    def eval_env(self, sellers):
        """
        The Consumer determines who, of those who sell the good he desires,
        he will buy from.

        Args:
            n_census: a list of stores selling the agent's desired good
        Returns:
            a store selling that good
        """
        top_seller = None
        max_util = 0.0
        for seller in sellers:
            this_util = seller.utils_from_good(self.goal)
            if this_util > max_util:
                max_util = this_util
                top_seller = seller
        self.last_utils = max_util
        return top_seller

    def respond_to_cond(self, store):
        """
        The agent moves to a store and buys from it.

        Args:
            store: the store to which we should move and buy from
        Returns:
            None
        """
        if store is not None:
            # print(self.name + " heading to " + store.name)
            self.move(store)
            store.purchase(self.allowance)

    def move(self, store):
        """
        Moves as close as possible to the store.

        Arg:
            store: retailer to which we move
        """
        open_spot = self.env.free_spot_near(store)
        if(open_spot is not None):
            self.env.move(self, open_spot[X], open_spot[Y])

    def postact(self):
        """
        We cycle through the good the agent wants each turn.
        """
        self.goal = (self.goal + 1) % NUM_GOODS


class Retailer(MarketParticipant):
    """
    A Retailer is a MarketParticipant, who not only sells goods, but who
    is also responsible for paying bills and maintianing sufficent capital
    for operation.

    Attributes:
        capital: If less than zero, the business disappears.
        expenses: how much is decremented from capital every step.
    """

    def __init__(self, name, goal, endowment, expenses, adj=MIN_ADJ):
        super().__init__(name, goal)
        self.capital = endowment
        self.expenses = expenses
        self.util_adj = adj

    def act(self):
        """
        Loses money. If it goes bankrupt, the business goes away.
        """
        self.env.user.tell(self.name + " has capital of: "
                           + str(self.capital)
                           + " and is paying expenses of "
                           + str(self.expenses))
        self.pay_bills(self.expenses)
        if(self.capital <= 0):
            self.declare_bankruptcy()

    def purchase(self, amt):
        """
        Agent buys from retailer, adding amt to retailer's capital.
        """
        self.capital += amt

    def declare_bankruptcy(self):
        """
        Removes the business from the town.
        """
        self.env.foreclose(self)

    def pay_bills(self, amt):
        """
        Lose capital.
        """
        self.capital -= amt

    def utils_from_good(self, good):
        if not self.sells(good):
            return NA
        else:
            return random.random() + self.util_adj

class MomAndPop(Retailer):
    """
    A small mom and pop store. It sells only one kind of good.

    Attributes:
        ntype: what it sells is its kind of store
    """

    def __init__(self, name, goal, endowment, expenses, adj=MIN_ADJ):
        # name them after what good they sell:
        type = GOODS_MAP[goal]
        super().__init__(type, goal, endowment, expenses, adj)
        self.ntype = type

    def sells(self, good):
        """
        Does this retailer sell this good?
        """
        return self.goal == good


class BigBox(Retailer):
    """
    A BigBox store. It sells all goods.
    """

    def __init__(self, name, goal, endowment, expenses):
        super().__init__(name, goal, endowment, expenses)

    def sells(self, good):
        """
        Does this retailer sell this good?
        The big box retailer sells everything!
        """
        return True


class EverytownUSA(ge.GridEnv):
    """
    Just your typical city: filled with businesses and consumers.
    The city management will remove businesses that cannot pay expenses!
    """

    def __init__(self, width, height, torus=False,
                 model_nm="Big Box Model", props=None):
        super().__init__(width=width, name=model_nm,
                         height=height, torus=torus,
                         model_nm=model_nm, postact=True,
                         props=props)
        self.utils = 0.0
        self.menu.view.add_menu_item("v", menu.MenuLeaf("(v)iew utility",
                                     self.view_util))
        self.menu.file.add_menu_item("u", menu.MenuLeaf("(u)tility report",
                                     self.report_util))
        self.add_variety("BigBox")
        self.mom_pop_pop = []
        self.util_hist = []

    def assemble_util_vars(self):
        varieties = None
        varieties = disp.assemble_lgraph_data("Mom and Pops", self.mom_pop_pop,
                                         disp.MAGENTA)
        varieties = disp.assemble_lgraph_data("Consumer Utils", self.util_hist,
                                         disp.GREEN, data=varieties)
        varieties = disp.assemble_lgraph_data("Big Boxes",
                                         self.agents.get_var_pop_hist("BigBox"),
                                         disp.BLUE, data=varieties)
        return varieties

    def view_util(self):
        """
        Graph population of stores versus consumer utility.
        """
        if self.period < 4:
            self.user.tell("Too little data to display")
            return

        self.line_graph = disp.LineGraph("Consumer Utility vs. # Retailers",
                                         self.assemble_util_vars(),
                                         self.period)

    def report_util(self):
        """
        Write CSV file with utility versus retail pop data.
        """
        file_nm = self.user.ask("Choose file name: ")
        varieties = self.assemble_util_vars()
        data.pop_report(file_nm, varieties)

    def postact_loop(self):
        """
        After all postactions, we add a BigBox store once we reach the 
        period the User determines during runtime.
        """
        super().postact_loop()
        # add big box store if right time.
        if self.period == self.props.get("bb_start_period"):
            self.add_agent(BigBox("Big Box Store",
                        goal="Dominance",
                        endowment=(self.props.get("endowment") * BB_MULT),
                        expenses=((self.props.get("expenses") * BB_MULT)
                              // BB_DIV)))
            # div BB_DIV sets the expense / capital ratio

    def foreclose(self, business):
        """
        Removes business from town.
        """
        self.remove_agent(business)

    def census(self, disp=True):
        """
        Take a census of our pops: here we want consumer util
        graphed against numbers of other types.
        """
        super().census(exclude_var="Consumer") # usual census for retailers
        utils = 0
        for shopper in self.variety_iter("Consumer"):
            utils += shopper.last_utils
        self.util_hist.append(utils)
        self.user.tell("Consumer utility gained this period: "
                       + str(int(utils)))
        mp_count = 0
        for agent in self.agents:
            if isinstance(agent, MomAndPop):
                mp_count += 1
        self.mom_pop_pop.append(mp_count)



from setuptools import setup, find_packages

setup(name='indras_net',
    version='1.3.0',
    description='A framework for agent-based modeling in Python.',
    url='https://github.com/gcallah/indras_net.git',
    author='Gene Callahan and Nathan Conroy',
    author_email='gcallah@mac.com',
    license='GNU',
    zip_safe=False,
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*",
                                 "tests"]),
    install_requires=[
          "networkx",
          "numpy",
    ],
    test_suite="",
    entry_points={
        "console_scripts": ['indra = indra.__main__:main']
    },
    classifiers=[
            # How mature is this project? Common values are
            #   3 - Alpha
            #   4 - Beta
            #   5 - Production/Stable
            'Development Status :: 4 - Beta',

            # Indicate who your project is intended for
            'Intended Audience :: Developers',

            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
    ],
)



"""
prop_args.py
Set, read, and write program-wide properties in one
location. Includes logging.
"""

import sys
import logging
import platform
import networkx as nx
import json
import indra.node as node
import indra.user as user

SWITCH = '-'

def in_range(low, val, high):
    return low <= val and val <= high


class PropArgs(node.Node):

    """
    This class holds sets of named properties for program-wide values.
    It enables getting properties from a file, in-program,
    or from the user, either via the command line or a prompt.
    """
    prop_sets = {}

    @staticmethod
    def get_props(model_nm):
        """
        Get properties for model 'model_nm'.
        """
        if model_nm in PropArgs.prop_sets:
            return PropArgs.prop_sets[model_nm]

    @staticmethod
    def create_props(model_nm, props={}):
        """
        Create a property object with values in 'props'.
        """
        return PropArgs(model_nm, props=props)

    @staticmethod
    def read_props(model_nm, file_nm):
        """
        Create a new PropArgs object from a json file
        """
        props = json.load(open(file_nm))
        return PropArgs.create_props(model_nm, props=props)

    def __init__(self, model_nm, logfile=None, props=None,
                 loglevel=logging.INFO):
        super().__init__("Properties")
        self.model_nm = model_nm
        # store this instance as the value in the dict for 'model_nm'
        PropArgs.prop_sets[model_nm] = self
        self.graph = nx.Graph()
        if props is None:
            self.props = {}
        else:
            self.props = props
            logfile = self.get("log_fname")
        self.logger = Logger(self, logfile=logfile)
        self.graph.add_edge(self, self.logger)
        self.set("OS", platform.system())
        self.set("model", model_nm)
        # process command line args and set them as properties:
        prop_nm = None
        for arg in sys.argv:
            if arg.startswith(SWITCH):
                prop_nm = arg.lstrip(SWITCH)
            elif prop_nm is not None:
                self.set(prop_nm, arg)
                prop_nm = None

    def display(self):
        """
        How to represent the properties on screen.
        """
        ret = "Properties for " + self.model_nm + "\n"
        for prop in self.props:
            ret += "\t" + prop + ": " + str(self.props[prop]) + "\n"

        return ret

    def set(self, nm, val):
        """
        Set a property value.
        """
        self.props[nm] = val

    def get(self, nm, default=None):
        """
        Get a property value, with a default
        that gets stored if the property is not there
        at the time of the call.
        """
        if nm not in self.props:
            self.props[nm] = default
        return self.props[nm]

    def ask(self, nm, msg, val_type, default=None, limits=None):
        """
        Ask the user for a property value, which might come
        from the user via the command line.
        """
        rng_msg = ""
        if limits is not None:
            (low, high) = limits
            rng_msg = " [" + str(low) + "-" + str(high) + "]"
        if default is not None:
            msg += " (" + str(default) + ")"
        msg += rng_msg + " "
        good_val = False
        while not good_val:
            if nm not in self.props:
                val = user.ask(msg)
                if len(val) == 0:
                    val = default
            else:  # was set from command line, but we still need to type it
                val = self.get(nm)
            typed_val = val_type(val)
            if limits is not None:
                good_val = in_range(low, typed_val, high)
            else:
                good_val = True
        self.set(nm, typed_val)

    def get_logfile(self):
        """
        Special get function for logfile name
        """
        return self.props.get("log_fname")

    def write(self, file_nm):
        """
        Write properties to json file.
        Useful for storing interesting parameter sets.
        """
        json.dump(self.props, open(file_nm, 'w'), indent=4)


class Logger(node.Node):
    """
    A class to track how we are logging.
    """

    DEF_FORMAT = '%(asctime)s:%(levelname)s:%(message)s'
    DEF_LEVEL = logging.INFO
    DEF_FILEMODE = 'w'
    DEF_FILENAME = 'log.txt'

    def __init__(self, props, logfile=None,
                 loglevel=logging.INFO):
        super().__init__("Logger")
        if logfile is None:
            logfile = Logger.DEF_FILENAME
        fmt = props.get("log_format", Logger.DEF_FORMAT)
        lvl = props.get("log_level", Logger.DEF_LEVEL)
        fmd = props.get("log_fmode", Logger.DEF_FILEMODE)
        fnm = props.get("log_fname", logfile)
        logging.basicConfig(format=fmt,
                            level=lvl,
                            filemode=fmd,
                            filename=fnm)
        logging.info("Logging initialized.")
